from tensorflow.keras.initializers import Initializer

from libspn_keras.initializers.dirichlet import Dirichlet

_DEFAULT_ACCUMULATOR_INITIALIZER = Dirichlet(alpha=1.0, axis=-2)


def set_default_accumulator_initializer(initializer: Initializer) -> None:
    """
    Configure the default accumulator that will be used for sum accumulators.

    Args:
        initializer: The initializer which will be used by default for sum accumulators.
    """
    global _DEFAULT_ACCUMULATOR_INITIALIZER
    _DEFAULT_ACCUMULATOR_INITIALIZER = initializer


def get_default_accumulator_initializer() -> Initializer:
    """
    Obtain default accumulator initializer.

    Returns:
        The default accumulator initializer that will be use in sum accumulators, unless specified
        explicitly at initialization.
    """
    global _DEFAULT_ACCUMULATOR_INITIALIZER
    return _DEFAULT_ACCUMULATOR_INITIALIZER



from libspn_keras.sum_ops import SumOpBase, SumOpGradBackprop

_DEFAULT_SUM_OP = None


def get_default_sum_op() -> SumOpBase:
    """
    Get default sum op.

    Returns:
        A ``SumOpBase`` instance that was set with ``set_default_sum_op``
    """
    return SumOpGradBackprop() if _DEFAULT_SUM_OP is None else _DEFAULT_SUM_OP


def set_default_sum_op(op: SumOpBase) -> None:
    """
    Set default sum op to conveniently use it throughout an SPN architecture.

    Args:
        op (SumOpBase): Implementation of sum op with corresponding backward pass definitions
    """
    global _DEFAULT_SUM_OP
    _DEFAULT_SUM_OP = op



from typing import Dict, Tuple, Union

import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.engine import data_adapter


class SumProductNetwork(keras.Model):
    """
    An SPN analogue of tensorflow.keras.Model that can be trained generatively.

    It does not expect labels y when calling .fit() if ``unsupervised`` == True.

    Args:
        unsupervised (bool): If ``True`` (default) the model does not expect label inputs in
            .fit() or .evaluate(). Also, losses and metrics should not expect a target output,
            just a y_hat.
    """

    def __init__(self, *args, unsupervised: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self.unsupervised = unsupervised

    def _train_step_unsupervised(
        self, data: Union[tf.Tensor, Tuple[tf.Tensor, ...]]
    ) -> Dict[str, tf.Tensor]:
        x, sample_weight, _ = data_adapter.unpack_x_y_sample_weight(data)
        with tf.GradientTape() as tape:
            out = self(x, training=True)
            dummy_target = tf.stop_gradient(out)
            loss = self.compiled_loss(
                dummy_target, out, sample_weight, regularization_losses=self.losses
            )

        trainable_variables = self.trainable_variables
        gradients = tape.gradient(loss, trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, trainable_variables))

        self.compiled_metrics.update_state(dummy_target, out, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    def _test_step_unsupervised(
        self, data: Union[tf.Tensor, Tuple[tf.Tensor, ...]]
    ) -> Dict[str, tf.Tensor]:
        x, sample_weight, _ = data_adapter.unpack_x_y_sample_weight(data)
        out = self(x, training=False)
        # Updates stateful loss metrics.
        dummy_target = tf.stop_gradient(out)
        self.compiled_loss(
            dummy_target, out, sample_weight, regularization_losses=self.losses
        )

        self.compiled_metrics.update_state(dummy_target, out, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    def train_step(
        self, data: Union[Tuple[tf.Tensor, ...], tf.Tensor]
    ) -> Dict[str, tf.Tensor]:
        """
        Train for one step.

        Args:
            data: Nested structure of tensors.

        Returns:
            Dict of metrics.
        """
        if self.unsupervised:
            return self._train_step_unsupervised(data)
        else:
            return super(SumProductNetwork, self).train_step(data)

    def test_step(
        self, data: Union[tf.Tensor, Tuple[tf.Tensor, ...]]
    ) -> Dict[str, tf.Tensor]:
        """
        Test for one step.

        Args:
            data: Nested structure of tensors.

        Returns:
            Dict of metrics.
        """
        if self.unsupervised:
            return self._test_step_unsupervised(data)
        else:
            return super(SumProductNetwork, self).test_step(data)



from typing import Dict, Tuple

import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.engine import data_adapter

from libspn_keras.layers.temporal_dense_product import TemporalDenseProduct


class DynamicSumProductNetwork(keras.Model):
    """
    SPN that re-uses its nodes at each time step.

    The input is expected to be pre-padded sequences with a full tensor shape
    of [num_batch, max_sequence_len, num_variables].

    Args:
        template_network: Template network that is applied to the leaves and ends with nodes that
            cover all variables for each timestep.
        interface_network_t0: Interface network for t = t0, applied on top of the template network's
            output at the current timestep.
        interface_network_t_minus_1: Interface network for t = t0 - 1, applied to the output of the
            interfaced output of the previous timestep
        top_network: Network on top of the interfaced network at the current timestep (covers all
            variables of the current timestep, including those of previous timesteps). This network
            must end with a root sum layer.
        return_last_step (bool): Whether to return only the roots at the last step with shape
            [num_batch, root_num_out] or whether to [num_batch, max_sequence_len, root_num_out]
        unsupervised (bool):
    """

    def __init__(
        self,
        template_network: keras.Model,
        interface_network_t0: keras.Model,
        interface_network_t_minus_1: keras.Model,
        top_network: keras.Model,
        return_last_step: bool = True,
        unsupervised: bool = True,
        **kwargs,
    ):
        super(DynamicSumProductNetwork, self).__init__(**kwargs)
        self.template_network = template_network
        self.unsupervised = unsupervised
        self.top_network = top_network
        self.interface_network_t0 = interface_network_t0
        self.interface_network_t_minus_1 = interface_network_t_minus_1
        self.return_last_step = return_last_step
        self.temporal_product = TemporalDenseProduct()

    def _train_step_unsupervised(self, data: Tuple[tf.Tensor]) -> Dict[str, tf.Tensor]:
        x, sequence_lens, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
        with tf.GradientTape() as tape:
            out = self([x, sequence_lens], training=True)
            dummy_target = tf.stop_gradient(out)
            loss = self.compiled_loss(
                dummy_target, out, sample_weight, regularization_losses=self.losses
            )

        trainable_variables = self.trainable_variables
        gradients = tape.gradient(loss, trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, trainable_variables))

        self.compiled_metrics.update_state(dummy_target, out, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    def _test_step_unsupervised(self, data: Tuple[tf.Tensor]) -> Dict[str, tf.Tensor]:
        x, sequence_lens, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
        out = self([x, sequence_lens], training=False)
        # Updates stateful loss metrics.
        dummy_target = tf.stop_gradient(out)
        self.compiled_loss(
            dummy_target, out, sample_weight, regularization_losses=self.losses
        )

        self.compiled_metrics.update_state(dummy_target, out, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    @tf.function
    def call(self, input_data: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Compute forward pass of the SPN.

        Args:
            input_data: Tuple of data Tensor and sequence length Tensor.

        Returns:
            Log probability of the root at each time step or that of the last timestep.

        Raises:
            ValueError: If number of input Tensors does not equal 2.
        """
        if len(input_data) != 2:
            raise ValueError(
                f"Dynamic SPN must be fed with data and sequence lengths tensors, "
                f"now got {len(input_data)} tensors."
            )
        input_data, sequence_lens = input_data[0], input_data[1]
        input_data = tf.transpose(input_data, [1, 0, 2])
        batch_size = tf.shape(input_data)[0]
        outputs = tf.TensorArray(tf.float32, batch_size)

        interface_network_t_minus_1_shape = tf.concat(
            [
                [tf.shape(input_data)[1]],
                [1, 1, self.interface_network_t0.output_shape[-1]],
            ],
            axis=0,
        )
        interface_t_minus_1 = tf.zeros(interface_network_t_minus_1_shape)
        for i in tf.range(batch_size):
            step_mask = tf.cast(
                tf.greater_equal(i, batch_size - tf.cast(sequence_lens, tf.int32)),
                tf.float32,
            )
            step_mask_flat = step_mask
            step_mask = tf.reshape(
                step_mask, tf.concat([[tf.shape(input_data)[1]], [1, 1, 1]], axis=0)
            )

            template_out = self.template_network(input_data[i]) * step_mask

            interface_t_minus_1 = interface_t_minus_1 * step_mask

            interface_t0 = self.interface_network_t0(template_out) * step_mask
            interface_template_prod = (
                self.temporal_product([interface_t_minus_1, interface_t0]) * step_mask
            )

            output = self.top_network(interface_template_prod) * tf.expand_dims(
                step_mask_flat, axis=1
            )
            outputs = outputs.write(i, output)

            interface_t_minus_1 = self.interface_network_t_minus_1(
                interface_template_prod
            )

        output = tf.transpose(outputs.stack(), [1, 0, 2])
        if self.return_last_step:
            output = output[:, -1, :]
        return output

    def train_step(self, data: Tuple[tf.Tensor]) -> Dict[str, tf.Tensor]:
        """
        Train for one step.

        Args:
            data: Nested structure of tensors.

        Returns:
            Dict of metrics.
        """
        if self.unsupervised:
            return self._train_step_unsupervised(data)
        else:
            return super(DynamicSumProductNetwork, self).train_step(data)

    def test_step(self, data: Tuple[tf.Tensor]) -> Dict[str, tf.Tensor]:
        """
        Test for one step.

        Args:
            data: Nested structure of tensors.

        Returns:
            Dict of metrics.
        """
        if self.unsupervised:
            return self._test_step_unsupervised(data)
        else:
            return super(DynamicSumProductNetwork, self).test_step(data)



from libspn_keras.losses.negative_log_joint import NegativeLogJoint
from libspn_keras.losses.negative_log_likelihood import NegativeLogLikelihood

__all__ = ["NegativeLogJoint", "NegativeLogLikelihood"]



from typing import Any, Callable, Tuple

import tensorflow as tf
import tensorflow_probability as tfp


class LocationScaleEMGradWrapper:
    """
    Wraps a location-scale distribution.

    As a consequence gradients that are passed down to the locations and scales result in EM updates.

    Args:
        location_scale_distribution: Distribution to wrap.
        first_order_moment_denom_accum: Denominator accumulator of first order moment.
        first_order_moment_num_accum: Numerator accumulator of first order moment.
        second_order_moment_denom_accum: Denominator accumulator of second order moment.
        second_order_moment_num_accum: Numerator accumulator of first order moment.
    """

    def __init__(
        self,
        location_scale_distribution: tfp.distributions.Distribution,
        first_order_moment_denom_accum: tf.Tensor,
        first_order_moment_num_accum: tf.Tensor,
        second_order_moment_denom_accum: tf.Tensor,
        second_order_moment_num_accum: tf.Tensor,
    ):
        self.location_scale_distribution = location_scale_distribution
        self.first_order_moment_denom_accum = first_order_moment_denom_accum
        self.first_order_moment_num_accum = first_order_moment_num_accum
        self.second_order_moment_denom_accum = second_order_moment_denom_accum
        self.second_order_moment_num_accum = second_order_moment_num_accum

    def __getattr__(self, attr: Any) -> Any:
        # see if this object has attr
        # NOTE do not use hasattr, it goes into
        # infinite recursion
        if attr in self.__dict__:
            # this object has it
            return getattr(self, attr)

        # proxy to the wrapped object
        return getattr(self.location_scale_distribution, attr)

    def log_prob(
        self, x: tf.Tensor
    ) -> Tuple[
        tf.Tensor,
        Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]],
    ]:
        """
        Compute log probability of the distribution at x.

        Args:
            x: The raw input data.

        Returns:
            A Tensor with the log probablity.
        """

        @tf.custom_gradient
        def _inner(
            first_order_moment_denom_accum: tf.Tensor,
            first_order_moment_num_accum: tf.Tensor,
            second_order_moment_denom_accum: tf.Tensor,
            second_order_moment_num_accum: tf.Tensor,
        ) -> Tuple[
            tf.Tensor,
            Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]],
        ]:
            def grad(
                dy: tf.Tensor,
            ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
                denom_grad = tf.reduce_sum(dy, axis=0, keepdims=True)
                first_order_moment_num_grad = tf.reduce_sum(
                    x * dy, axis=0, keepdims=True
                )
                second_order_moment_num_grad = tf.reduce_sum(
                    tf.square(x) * dy, axis=0, keepdims=True
                )

                return (
                    denom_grad,
                    first_order_moment_num_grad,
                    denom_grad,
                    second_order_moment_num_grad,
                )

            out = self.location_scale_distribution.log_prob(x)

            return out, grad

        return _inner(
            self.first_order_moment_denom_accum,
            self.first_order_moment_num_accum,
            self.second_order_moment_denom_accum,
            self.second_order_moment_num_accum,
        )


class LocationEMGradWrapper:
    """
    Wraps a location-scale distribution.

    As a consequence gradients that are passed down to the locations result in EM updates.

    Args:
        location_scale_distribution: Distribution to wrap.
        first_order_moment_denom_accum: Denominator accumulator of first order moment.
        first_order_moment_num_accum: Numerator accumulator of first order moment.
    """

    def __init__(
        self,
        location_scale_distribution: tfp.distributions.Distribution,
        first_order_moment_denom_accum: tf.Tensor,
        first_order_moment_num_accum: tf.Tensor,
    ):
        self.location_scale_distribution = location_scale_distribution
        self.first_order_moment_denom_accum = first_order_moment_denom_accum
        self.first_order_moment_num_accum = first_order_moment_num_accum

    def __getattr__(self, attr: Any) -> Any:
        # see if this object has attr
        # NOTE do not use hasattr, it goes into
        # infinite recursion
        if attr in self.__dict__:
            # this object has it
            return getattr(self, attr)

        # proxy to the wrapped object
        return getattr(self.location_scale_distribution, attr)

    def log_prob(
        self, x: tf.Tensor
    ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:
        """
        Compute log probabilty of the distribution at x.

        Args:
            x: The raw input data.

        Returns:
            A Tensor with the log probablity.
        """

        @tf.custom_gradient
        def _inner(
            first_order_moment_denom_accum: tf.Tensor,
            first_order_moment_num_accum: tf.Tensor,
        ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:
            def grad(dy: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
                denom_grad = tf.reduce_sum(dy, axis=0, keepdims=True)
                first_order_moment_num_grad = tf.reduce_sum(
                    x * dy, axis=0, keepdims=True
                )

                return denom_grad, first_order_moment_num_grad

            out = self.location_scale_distribution.log_prob(x)

            return out, grad

        return _inner(
            self.first_order_moment_denom_accum, self.first_order_moment_num_accum
        )



import logging
from typing import List, Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow import initializers
from tensorflow import keras


logger = logging.getLogger("libspn-keras")


class Conv2DProduct(keras.layers.Layer):
    """
    Convolutional product as described in (Van de Wolfshaar and Pronobis, 2019).

    Expects log-space inputs and produces log-space outputs.

    Args:
        strides: A tuple or list of strides
        dilations: A tuple or list of dilations
        kernel_size: A tuple or list of kernel sizes
        num_channels: Number of channels. If None, will be set to
            num_in_channels ** prod(kernel_sizes). This can be source of OOM problems quickly.
        padding: Can be either 'full', 'valid' or 'final'. Use 'final' for the top ConvProduct
            of a DGC-SPN. The other choices have the standard interpretation. Valid padding
            usually requires non-overlapping patches, whilst full padding is used with
            overlapping patches and expontentially increasing dilation rates, see also
            [Van de Wolfshaar, Pronobis (2019)].
        depthwise: Whether to use depthwise convolutions. If True, the value of num_channels
            will be ignored
        **kwargs: Keyword arguments to pass on to the keras.Layer superclass.

    References:
        Deep Generalized Convolutional Sum-Product Networks for Probabilistic Image Representations,
        `Van de Wolfshaar, Pronobis (2019) <https://arxiv.org/abs/1902.06155>`_
    """

    def __init__(
        self,
        strides: List[int],
        dilations: List[int],
        kernel_size: List[int],
        num_channels: Optional[int] = None,
        padding: str = "valid",
        depthwise: bool = False,
        **kwargs
    ):
        super(Conv2DProduct, self).__init__(**kwargs)
        self.strides = strides
        self.dilations = dilations
        self.num_channels = num_channels
        self.padding = padding
        self.kernel_size = kernel_size
        self.depthwise = depthwise

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Construct inner components of the layer.

        Args:
            input_shape: Input shape of the layer.
        """
        self._build_depthwise(
            input_shape
        ) if self.depthwise else self._build_onehot_kernels(input_shape)
        super(Conv2DProduct, self).build(input_shape)

    def _build_onehot_kernels(self, input_shape: Tuple[Optional[int], ...]) -> None:
        _, num_scopes_vertical, num_scopes_horizontal, num_channels_in = input_shape

        if num_scopes_vertical is None:
            raise ValueError("Cannot build Conv2DProduct: unknown vertical dimension")
        if num_scopes_horizontal is None:
            raise ValueError("Cannot build Conv2DProduct: unknown horizontal dimension")
        if num_channels_in is None:
            raise ValueError("Cannot build Conv2DProduct: unknown channel dimension")

        self._spatial_dim_sizes = num_scopes_vertical, num_scopes_horizontal

        sparse_kernels = self._create_sparse_kernels(num_channels_in, self.num_channels)

        onehot_kernels = self._sparse_kernels_to_onehot(sparse_kernels, num_channels_in)

        self._onehot_kernels = self.add_weight(
            "onehot_kernel",
            initializer=initializers.Constant(onehot_kernels),
            trainable=False,
            shape=onehot_kernels.shape,
        )

    def _build_depthwise(self, input_shape: Tuple[Optional[int], ...]) -> None:
        _, num_scopes_vertical, num_scopes_horizontal, num_channels_in = input_shape

        if num_scopes_vertical is None:
            raise ValueError("Cannot build Conv2DProduct: unknown vertical dimension")
        if num_scopes_horizontal is None:
            raise ValueError("Cannot build Conv2DProduct: unknown horizontal dimension")
        if num_channels_in is None:
            raise ValueError("Cannot build Conv2DProduct: unknown channel dimension")

        self._spatial_dim_sizes = num_scopes_vertical, num_scopes_horizontal
        self.num_channels = num_channels_in

        sparse_kernels = self._create_sparse_kernels(1, 1)

        onehot_kernels = self._sparse_kernels_to_onehot(sparse_kernels, 1)

        self._onehot_kernels = self.add_weight(
            "onehot_kernel",
            initializer=initializers.Constant(onehot_kernels),
            trainable=False,
            shape=onehot_kernels.shape,
        )

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute the 2D convolutional log-product (CLP).

        Args:
            x: Spatial Tensor.
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with local products of the input.
        """
        if self.depthwise:
            return self._call_depthwise(x)
        else:
            return self._call_onehot_kernels(x)

    def _call_onehot_kernels(self, x: tf.Tensor) -> tf.Tensor:
        # Split in list of tensors which will be added up using outer products
        with tf.name_scope("OneHotConv"):
            pad_left, pad_right, pad_top, pad_bottom = self._pad_sizes()

            x_padded = tf.pad(
                x, [[0, 0], [pad_top, pad_bottom], [pad_left, pad_right], [0, 0]]
            )
            out = tf.nn.conv2d(
                x_padded,
                self._onehot_kernels,
                strides=self.strides,
                padding="VALID",
                dilations=self.dilations,
            )
            return out

    def _call_depthwise(self, x: tf.Tensor) -> tf.Tensor:
        # Split in list of tensors which will be added up using outer products
        with tf.name_scope("DepthwiseConv"):
            pad_left, pad_right, pad_top, pad_bottom = self._pad_sizes()
            channels_first = tf.reshape(
                tf.transpose(x, (0, 3, 1, 2)), (-1,) + self._spatial_dim_sizes + (1,)
            )
            x_padded = tf.pad(
                channels_first,
                [[0, 0], [pad_top, pad_bottom], [pad_left, pad_right], [0, 0]],
            )
            out = tf.nn.conv2d(
                x_padded,
                self._onehot_kernels,
                strides=self.strides,
                padding="VALID",
                dilations=self.dilations,
            )

            spatial_dim_sizes_out = self._compute_out_size_spatial(
                *self._spatial_dim_sizes
            )

            return tf.transpose(
                tf.reshape(out, (-1, self.num_channels) + spatial_dim_sizes_out),
                (0, 2, 3, 1),
            )

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.

        Raises:
            ValueError: In case shape could not be determined.
        """
        num_batch, num_scopes_vertical_in, num_scopes_horizontal_in, _ = input_shape
        if num_scopes_vertical_in is None:
            raise ValueError("Cannot build Conv2DProduct: unknown vertical dimension")
        if num_scopes_horizontal_in is None:
            raise ValueError("Cannot build Conv2DProduct: unknown horizontal dimension")
        (
            num_scopes_vertical_out,
            num_scopes_horizontal_out,
        ) = self._compute_out_size_spatial(
            num_scopes_vertical_in, num_scopes_horizontal_in
        )
        return (
            num_batch,
            num_scopes_vertical_out,
            num_scopes_horizontal_out,
            self.num_channels,
        )

    def _sparse_kernels_to_onehot(
        self, sparse_kernels: np.ndarray, num_channels_in: int
    ) -> np.ndarray:
        """
        Convert an index-based representation of sparse kernels to a dense onehot representation.

        Args:
            sparse_kernels (numpy.ndarray): A sparse kernel representation of shape
                [rows, cols, output_channel] containing the indices for which the kernel equals 1.
            num_channels_in: Number of channels in input.

        Returns:
            A onehot representation of the same kernel with shape
            [rows, cols, input_channel, output_channel].
        """
        kernels_height, kernels_width, num_channels_out = sparse_kernels.shape
        onehot_kernels = np.ones(
            (kernels_height, kernels_width, num_channels_out, num_channels_in)
        )
        onehot_kernels *= np.arange(num_channels_in).reshape([1, 1, 1, num_channels_in])
        onehot_kernels = np.equal(
            onehot_kernels, np.expand_dims(sparse_kernels, -1)
        ).astype(np.float32)
        return np.transpose(onehot_kernels, (0, 1, 3, 2))

    def _create_sparse_kernels(
        self, num_channels_in: int, num_channels_out: Optional[int]
    ) -> np.ndarray:
        """
        Generate sparse kernels.

        These kernels only contain '1' on a single channel
        per row and column. The remaining values are all zero. This method returns the sparse
        representation, containing only the indices for which the kernels are 1 along the input
        channel axis.

        Args:
            num_channels_in: Number of input channels.
            num_channels_out (int): The number of output channels. In case the number of channels given is
                larger than the number of possible one-hot assignments, a warning is given and the
                number of channels is set accordingly before generating the connections.

        Returns:
            A `numpy.ndarray` containing the 'sparse' representation of the kernels with shape
            `[row, column, channel]`, containing the indices of the input channel for which the
            kernel is 1.
        """
        if num_channels_out is None:
            num_channels_out = self.num_channels = int(
                num_channels_in ** np.prod(self.kernel_size)
            )

        kernel_surface = int(np.prod(self.kernel_size))
        total_possibilities = int(num_channels_in ** kernel_surface)
        if num_channels_out >= total_possibilities:
            if num_channels_out > total_possibilities:
                logger.warning(
                    "Number of channels exceeds total number of combinations."
                )
                self.num_channels = total_possibilities
            p = np.arange(total_possibilities)
            kernel_cells = []
            for _ in range(kernel_surface):
                kernel_cells.append(p % num_channels_in)
                p //= num_channels_in
            return np.stack(kernel_cells, axis=0).reshape(
                self.kernel_size + [total_possibilities]
            )

        if num_channels_out >= num_channels_in:
            kernel_cells = []
            for _ in range(kernel_surface):
                ind = np.arange(num_channels_out) % num_channels_in
                np.random.shuffle(ind)
                kernel_cells.append(ind)
            return np.asarray(kernel_cells).reshape(
                [*self.kernel_size, num_channels_out]
            )

        sparse_shape = self.kernel_size + [num_channels_out]
        size = int(np.prod(sparse_shape))
        return np.random.randint(num_channels_in, size=size).reshape(sparse_shape)

    def _effective_kernel_size(self) -> List[int]:
        """
        Compute the 'effective' kernel size by also taking into account the dilation rate.

        Returns:
            tuple: A tuple with (num_kernel_rows, num_kernel_cols)
        """
        kernel_sizes = [
            (ks - 1) * d + 1 for ks, d in zip(self.kernel_size, self.dilations)
        ]
        return kernel_sizes

    def _pad_sizes(self) -> Tuple[int, ...]:
        """
        Determine the pad sizes.

        Returns:
            A tuple of left, right, top and bottom padding sizes.

        Raises:
            ValueError: When padding algorithm is unknown.
        """
        if self.padding == "valid":
            return 0, 0, 0, 0
        if self.padding == "full":
            kernel_height, kernel_width = self._effective_kernel_size()
            pad_top = pad_bottom = kernel_height - 1
            pad_left = pad_right = kernel_width - 1
            return pad_left, pad_right, pad_top, pad_bottom
        if self.padding == "final":
            kernel_height, kernel_width = self._effective_kernel_size()
            pad_top = (
                (kernel_height - 1) * 2 - self._spatial_dim_sizes[0]
                if self.kernel_size[0] > 1
                else 0
            )
            pad_left = (
                (kernel_width - 1) * 2 - self._spatial_dim_sizes[1]
                if self.kernel_size[1] > 1
                else 0
            )
            return pad_left, 0, pad_top, 0
        raise ValueError(
            "{}: invalid padding algorithm. Use 'valid', 'full' or 'wicker_top', got '{}'".format(
                self, self.padding
            )
        )

    def _compute_out_size_spatial(
        self, num_scopes_vertical_in: int, num_scopes_horizontal_in: int
    ) -> Tuple[int, ...]:
        """
        Compute spatial output shape.

        Args:
            num_scopes_vertical_in: Number of scopes on vertical axis.
            num_scopes_horizontal_in: Number of scopes on horizontal axis.

        Returns:
            A tuple with (num_rows, num_cols, num_channels).
        """
        kernel_size0, kernel_size1 = self._effective_kernel_size()

        pad_left, pad_right, pad_top, pad_bottom = self._pad_sizes()

        rows_post_pad = pad_top + pad_bottom + num_scopes_vertical_in - kernel_size0 + 1
        cols_post_pad = (
            pad_left + pad_right + num_scopes_horizontal_in - kernel_size1 + 1
        )
        return tuple(
            int(np.ceil(post_pad / s))
            for post_pad, s in zip([rows_post_pad, cols_post_pad], self.strides)
        )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(
            strides=self.strides,
            dilations=self.dilations,
            num_channels=self.num_channels,
            padding=self.padding,
            kernel_size=self.kernel_size,
            depthwise=self.depthwise,
        )
        base_config = super(Conv2DProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Dict, List, Optional, Tuple, Union

import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.engine import data_adapter
from tensorflow.python.keras.engine.sequential import (
    _get_shape_tuple,
    SINGLE_LAYER_OUTPUT_ERROR_MSG,
)
from tensorflow.python.util import nest

from libspn_keras.layers.base_leaf import BaseLeaf
from libspn_keras.layers.dense_product import DenseProduct
from libspn_keras.layers.dense_sum import DenseSum
from libspn_keras.layers.flat_to_regions import FlatToRegions
from libspn_keras.layers.location_scale_leaf import LocationScaleLeafBase
from libspn_keras.layers.log_dropout import LogDropout
from libspn_keras.layers.normalize_standard_score import NormalizeStandardScore
from libspn_keras.layers.permute_and_pad_scopes_random import PermuteAndPadScopes
from libspn_keras.layers.permute_and_pad_scopes_random import PermuteAndPadScopesRandom
from libspn_keras.layers.reduce_product import ReduceProduct
from libspn_keras.layers.root_sum import RootSum
from libspn_keras.layers.undecompose import Undecompose


class SequentialSumProductNetwork(keras.Sequential):
    """
    An analogue of ``tensorflow.keras.Sequential`` that can be trained in an unsupervised way.

    It does not expect labels y when calling .fit() if ``unsupervised`` == True. Inherits from
    ``keras.Sequential``, so layers are passed to it as a list.

    Args:
        unsupervised (bool): If ``True`` the model does not expect label inputs in .fit() or
            .evaluate(). Also, losses and metrics should not expect a target output, just a y_hat.
            By default, it will be inferred from ``infer_no_evidence`` and otherwise defaults to
            ``True``.
        infer_no_evidence (bool): If ``True``, the model expects an evidence mask defined as a
            boolean tensor which is used to mask out variables that are not part of the evidence.
    """

    def __init__(
        self,
        layers: List[tf.keras.layers.Layer],
        infer_no_evidence: bool = False,
        unsupervised: Optional[bool] = None,
        **kwargs
    ):
        self._infer_factors_for_region_spn(layers)
        if unsupervised is None:
            unsupervised = False if infer_no_evidence else True
        super().__init__(layers, **kwargs)
        self.unsupervised = unsupervised
        self.infer_no_evidence = infer_no_evidence
        if infer_no_evidence and unsupervised:
            raise ValueError(
                "Model cannot be unsupervised when evidence should be inferred"
            )
        self.infer_no_evidence = infer_no_evidence
        if infer_no_evidence:
            self._normalize_index = self._normalize_layer = None
            for i, layer in enumerate(self.layers):
                if isinstance(layer, NormalizeStandardScore):
                    self._normalize_index = i
                    self._normalize_layer = layer

                if isinstance(layer, LocationScaleLeafBase):
                    self._leaf_index = i
                    self._leaf_layer = layer
                    break
            else:
                raise ValueError("No LocationScaleLeafBase leaf layer found")

    def call(
        self,
        inputs: Union[Tuple[tf.Tensor, ...], tf.Tensor],
        training: Optional[bool] = None,
        mask: Optional[tf.Tensor] = None,
    ) -> tf.Tensor:
        """
        Call the SPN.

        Optionally, the SPN will infer evidence depending on the chosen backprop in the sum nodes.

        Args:
            inputs: Nested structure of input tensors.
            training: Whether the SPN is trainin or not.
            mask: Mask to zero out values.

        Returns:
            Either the inferred values for missing evidence or the log-probability of the root.

        Raises:
            ValueError: When evidence should be inferred and there's no 2-tuple of Tensors provided.
        """
        if self.infer_no_evidence:
            if not isinstance(inputs, tuple) or len(inputs) != 2:
                raise ValueError(
                    "Expected tuple of Tensors with length 2 with data and evidence mask."
                )
            return self._call_backprop_to_leaves(inputs, training)
        return super(SequentialSumProductNetwork, self).call(inputs, training, mask)

    def _train_step_masked_leaves(self, data: tf.Tensor) -> Dict[str, tf.Tensor]:
        x, evidence_mask, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
        with tf.GradientTape() as tape:
            out = self([x, evidence_mask], training=True)
            no_evidence = tf.logical_not(evidence_mask)
            x_without_evidence = tf.boolean_mask(x, no_evidence)
            out_without_evidence = tf.boolean_mask(out, no_evidence)
            loss = self.compiled_loss(
                x_without_evidence,
                out_without_evidence,
                sample_weight,
                regularization_losses=self.losses,
            )

        trainable_variables = self.trainable_variables
        gradients = tape.gradient(loss, trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, trainable_variables))

        self.compiled_metrics.update_state(
            x_without_evidence, out_without_evidence, sample_weight
        )
        return {m.name: m.result() for m in self.metrics}

    def _call_backprop_to_leaves(
        self, inputs: Tuple[tf.Tensor, ...], training: Optional[bool] = None
    ) -> tf.Tensor:
        inputs, evidence_mask = inputs[0], inputs[1]
        if self._build_input_shape is None:  # type: ignore
            input_shapes = nest.map_structure(_get_shape_tuple, inputs)
            self._build_input_shape = input_shapes

        outputs = inputs  # handle the corner case where self.layers is empty
        with tf.GradientTape() as tape:
            for i, layer in enumerate(self.layers):
                # During each iteration, `inputs` are the inputs to `layer`, and `outputs`
                # are the outputs of `layer` applied to `inputs`. At the end of each
                # iteration `inputs` is set to `outputs` to prepare for the next layer.
                kwargs = {}
                argspec = self._layer_call_argspecs[layer].args
                if "training" in argspec:
                    kwargs["training"] = training

                if i == self._leaf_index:
                    leaf_inputs = inputs

                if i == self._normalize_index:
                    kwargs["return_stats"] = True
                    outputs, mean, stddev = layer(inputs, **kwargs)
                else:
                    outputs = layer(inputs, **kwargs)

                if i == self._leaf_index:
                    outputs = leaf_out = tf.where(
                        evidence_mask, outputs, tf.zeros_like(outputs)
                    )
                    tape.watch(leaf_out)

                if len(nest.flatten(outputs)) != 1:
                    raise ValueError(SINGLE_LAYER_OUTPUT_ERROR_MSG)
                # `outputs` will be the inputs to the next layer.
                inputs = outputs

        leaf_grads = tape.gradient(outputs, leaf_out)
        modes = self._leaf_layer.get_modes()
        outputs = tf.reduce_sum(tf.expand_dims(leaf_grads, axis=-1) * modes, axis=3)
        outputs = tf.where(evidence_mask, leaf_inputs, outputs)
        if self._normalize_layer is not None and self._normalize_index is not None:
            outputs = (
                outputs * (stddev + self._normalize_layer.normalization_epsilon) + mean
            )

        return outputs

    def _train_step_unsupervised(self, data: tf.Tensor) -> Dict[str, tf.Tensor]:
        x, sample_weight, _ = data_adapter.unpack_x_y_sample_weight(data)
        with tf.GradientTape() as tape:
            out = self(x, training=True)
            dummy_target = tf.stop_gradient(out)
            loss = self.compiled_loss(
                dummy_target, out, sample_weight, regularization_losses=self.losses
            )

        trainable_variables = self.trainable_variables
        gradients = tape.gradient(loss, trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, trainable_variables))

        self.compiled_metrics.update_state(dummy_target, out, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    def _test_step_unsupervised(self, data: tf.Tensor) -> Dict[str, tf.Tensor]:
        x, sample_weight, _ = data_adapter.unpack_x_y_sample_weight(data)
        out = self(x, training=False)
        # Updates stateful loss metrics.
        dummy_target = tf.stop_gradient(out)
        self.compiled_loss(
            dummy_target, out, sample_weight, regularization_losses=self.losses
        )

        self.compiled_metrics.update_state(dummy_target, out, sample_weight)
        return {m.name: m.result() for m in self.metrics}

    def train_step(self, data: tf.Tensor) -> Dict[str, tf.Tensor]:
        """
        Train for one step.

        Args:
            data: Nested structure of tensors.

        Returns:
            Dict of metrics.
        """
        if self.infer_no_evidence:
            return self._train_step_masked_leaves(data)
        elif self.unsupervised:
            return self._train_step_unsupervised(data)
        else:
            return super(SequentialSumProductNetwork, self).train_step(data)

    def test_step(self, data: tf.Tensor) -> Dict[str, tf.Tensor]:
        """
        Test for one step.

        Args:
            data: Nested structure of tensors.

        Returns:
            Dict of metrics.
        """
        if self.unsupervised:
            return self._test_step_unsupervised(data)
        else:
            return super(SequentialSumProductNetwork, self).test_step(data)

    @staticmethod
    def _is_region_spn(layers: List[tf.keras.layers.Layer]) -> bool:
        for layer in layers:
            if not isinstance(
                layer,
                (
                    DenseProduct,
                    DenseSum,
                    ReduceProduct,
                    RootSum,
                    PermuteAndPadScopes,
                    FlatToRegions,
                    LogDropout,
                    BaseLeaf,
                    NormalizeStandardScore,
                    Undecompose,
                ),
            ):
                return False

        return True

    def _infer_factors_for_region_spn(
        self, layers: List[tf.keras.layers.Layer]
    ) -> None:
        if self._is_region_spn(layers):
            for layer in layers:
                if isinstance(layer, PermuteAndPadScopesRandom):
                    if layer.factors is None:
                        layer.set_factors(
                            [
                                layer.num_factors
                                for layer in layers
                                if isinstance(layer, (DenseProduct, ReduceProduct))
                            ]
                        )



import tensorflow as tf

from libspn_keras.math.logutils import replace_infs_with_zeros


def logconv1x1_2d(input: tf.Tensor, filter: tf.Tensor) -> tf.Tensor:
    r"""
    Convolution in logspace with 1x1 filters.

    Computes :math:`\log(\text{conv}(\mathtt{input},\mathtt{filter}))` from
    :math:`\log(\mathtt{input})` and :math:`\log(\mathtt{filter})`

    Args:
        input: Input in logspace
        filter: Filter in logspace

    Returns:
        Convolution of input and filter in logspace
    """
    with tf.name_scope("LogConv1x1"):
        filter_max = replace_infs_with_zeros(
            tf.stop_gradient(tf.reduce_max(filter, axis=-2, keepdims=True))
        )
        input_max = replace_infs_with_zeros(
            tf.stop_gradient(tf.reduce_max(input, axis=-1, keepdims=True))
        )

        out = tf.math.log(
            tf.nn.convolution(
                input=tf.exp(input - input_max),
                filters=tf.exp(filter - filter_max),
                padding="SAME",
            )
        )
        out += filter_max + input_max

        return out



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow.keras import initializers
from tensorflow_probability import distributions


class Dirichlet(initializers.Initializer):
    r"""
    Initializes all values in a tensor with :math:`Dir(\alpha)`.

    Args:
        axis: The axis over which to sample from a :math:`Dir(\alpha)`.
        alpha: The :math:`\alpha` parameter of the Dirichlet distribution.
            If a scalar, this is broadcast along the given axis.
    """

    def __init__(
        self, axis: int = -2, alpha: float = 0.1,
    ):
        self.axis = axis
        self.alpha = alpha

    def __call__(
        self, shape: Tuple[int, ...], dtype: Optional[tf.DType] = None
    ) -> tf.Tensor:
        """
        Compute initializations along the given axis.

        Args:
            shape: Shape of Tensor to initialize.
            dtype: DType of Tensor to initialize.

        Returns:
            Initial value.
        """
        axis = (len(shape) + self.axis) % len(shape)

        alpha_as_tensor = tf.convert_to_tensor(self.alpha)
        alpha = (
            tf.tile([alpha_as_tensor], [shape[axis]])
            if tf.size(alpha_as_tensor) == 1
            else alpha_as_tensor
        )
        dirichlet_sample = distributions.Dirichlet(concentration=alpha).sample(
            [dim for i, dim in enumerate(shape) if i != axis]
        )

        perm = [
            i if i < axis else len(shape) - 1 if i == axis else i - 1
            for i in range(len(shape))
        ]
        return tf.cast(tf.transpose(dirichlet_sample, perm), dtype)

    def get_config(self) -> dict:
        """
        Obtain the config.

        Returns:
            Key-value mapping with the configuration of this initializer.
        """
        return {"alpha": self.alpha, "axis": self.axis}



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import regularizers

from libspn_keras.config.accumulator_initializer import (
    get_default_accumulator_initializer,
)
from libspn_keras.config.linear_accumulator_constraint import (
    get_default_linear_accumulators_constraint,
)
from libspn_keras.config.logspace_accumulator_constraint import (
    get_default_logspace_accumulators_constraint,
)
from libspn_keras.config.sum_op import get_default_sum_op
from libspn_keras.constraints import LogNormalized
from libspn_keras.constraints.greater_equal_epsilon_normalized import (
    GreaterEqualEpsilonNormalized,
)
from libspn_keras.logspace import logspace_wrapper_initializer
from libspn_keras.sum_ops import SumOpBase


class DenseSum(keras.layers.Layer):
    """
    Computes densely connected sums per scope and decomposition.

    Expects incoming ``Tensor`` to be of shape [num_scopes, num_decomps, num_batch, num_nodes]. If your
    input is passed through a ``FlatToRegions`` layer this is already taken care of.

    Args:
        num_sums: Number of sums per scope
        logspace_accumulators: If ``True``, accumulators will be represented in log-space which
            is typically used with ``BackpropMode.GRADIENT``. If ``False``, accumulators will be
            represented in linear space. Weights are computed by normalizing the accumulators
            per sum, so that we always end up with a normalized SPN. If ``None`` (default) it
            will be set to ``True`` for ``BackpropMode.GRADIENT`` and ``False`` otherwise.
        accumulator_initializer: Initializer for accumulator. Will automatically be converted
            to log-space values if ``logspace_accumulators`` is enabled.
        accumulator_regularizer: Regularizer for accumulator (experimental)
        linear_accumulator_constraint: Constraint for accumulator defaults to constraint that
            ensures small positive constant at minimum. Will be ignored if logspace_accumulators
            is set to True.
        sum_op (SumOpBase): SumOpBase instance which determines how to compute the forward and
            backward pass of the weighted sums
        **kwargs: kwargs to pass on to keras.Layer super class
    """

    def __init__(
        self,
        num_sums: int,
        logspace_accumulators: Optional[bool] = None,
        accumulator_initializer: Optional[keras.initializers.Initializer] = None,
        sum_op: Optional[SumOpBase] = None,
        accumulator_regularizer: Optional[keras.regularizers.Regularizer] = None,
        logspace_accumulator_constraint: Optional[keras.constraints.Constraint] = None,
        linear_accumulator_constraint: Optional[keras.constraints.Constraint] = None,
        **kwargs
    ):
        super(DenseSum, self).__init__(**kwargs)
        self.num_sums = num_sums
        self.sum_op = sum_op or get_default_sum_op()
        self.logspace_accumulators = (
            self.sum_op.default_logspace_accumulators()
            if logspace_accumulators is None
            else logspace_accumulators
        )
        self.accumulator_initializer = (
            accumulator_initializer or get_default_accumulator_initializer()
        )
        self.accumulator_regularizer = accumulator_regularizer
        self.linear_accumulator_constraint = (
            linear_accumulator_constraint
            or get_default_linear_accumulators_constraint()
        )
        self.logspace_accumulator_constraint = (
            logspace_accumulator_constraint
            or get_default_logspace_accumulators_constraint()
        )

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.
        """
        # Create a trainable weight variable for this layer.
        _, self._num_scopes, self._num_decomps, self._num_nodes_in = input_shape

        weights_shape = (
            self._num_scopes,
            self._num_decomps,
            self._num_nodes_in,
            self.num_sums,
        )

        initializer = self.accumulator_initializer
        accumulator_constraint = self.linear_accumulator_constraint
        if self.logspace_accumulators:
            initializer = logspace_wrapper_initializer(self.accumulator_initializer)
            accumulator_constraint = self.logspace_accumulator_constraint

        self._accumulators = self.add_weight(
            name="sum_weights",
            shape=weights_shape,
            initializer=initializer,
            regularizer=self.accumulator_regularizer,
            constraint=accumulator_constraint,
        )
        if accumulator_constraint is not None:
            self._accumulators.assign(accumulator_constraint(self._accumulators))
        self._forward_normalize = not isinstance(
            accumulator_constraint, (GreaterEqualEpsilonNormalized, LogNormalized)
        )
        super(DenseSum, self).build(input_shape)

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute the probability of the leaf nodes.

        Args:
            x: Spatial or region Tensor with raw input values.
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with the probabilities per component.
        """
        return self.sum_op.weighted_sum(
            x, self._accumulators, self.logspace_accumulators, self._forward_normalize
        )

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.
        """
        num_scopes, num_decomps, num_batch, _ = input_shape
        return num_batch, num_scopes, num_decomps, self.num_sums

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(
            num_sums=self.num_sums,
            accumulator_initializer=initializers.serialize(
                self.accumulator_initializer
            ),
            logspace_accumulators=self.logspace_accumulators,
            accumulator_regularizer=regularizers.serialize(
                self.accumulator_regularizer
            ),
            linear_accumulator_constraint=constraints.serialize(
                self.linear_accumulator_constraint
            ),
        )
        base_config = super(DenseSum, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras


class SpatialToRegions(keras.layers.Layer):
    """
    Reshapes spatial SPN layer to a dense layer.

    The dense output has leading dimensions for scopes and decomps (which will be ``[1, 1]``).
    """

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: If dimensions are unknown.
        """
        _, num_cells_vertical, num_cells_horizontal, num_nodes = input_shape
        if num_cells_horizontal is None:
            raise ValueError(
                "Cannot compute shape with unknown number of horizontal cells"
            )
        if num_cells_vertical is None:
            raise ValueError(
                "Cannot compute shape with unknown number of vertical cells"
            )
        if num_nodes is None:
            raise ValueError("Cannot compute shape with unknown number of nodes")
        self._out_num_nodes = num_cells_vertical * num_cells_horizontal * num_nodes

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute region representation from spatial tensor.

        Assumes that all nodes have the same scope along the spatial axes.

        Args:
            x: Spatial input.
            kwargs: Remaining keyword arguments.

        Returns:
            A region representation of the input.
        """
        shape = tf.shape(x)
        return tf.reshape(x, [shape[0], 1, 1, self._out_num_nodes])

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.

        Raises:
            ValueError: When shape cannot be computed.
        """
        num_batch, num_cells_vertical, num_cells_horizontal, num_nodes = input_shape
        if num_cells_horizontal is None:
            raise ValueError(
                "Cannot compute shape with unknown number of horizontal cells"
            )
        if num_cells_vertical is None:
            raise ValueError(
                "Cannot compute shape with unknown number of vertical cells"
            )
        if num_nodes is None:
            raise ValueError("Cannot compute shape with unknown number of nodes")
        return num_batch, 1, 1, num_cells_vertical * num_cells_horizontal * num_nodes



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow_probability import distributions
from tensorflow_probability.python.internal.reparameterization import (
    NOT_REPARAMETERIZED,
)

from libspn_keras.layers.base_leaf import BaseLeaf


class IndicatorLeaf(BaseLeaf):
    """
    Indicator leaf distribution taking integer inputs and producing a discrete indicator representation.

    This effectively comes down to computing a one-hot representation along the final axis.

    Args:
        num_components (int): Number of components, or indicators in this context.
        dtype: Dtype of input
        **kwargs: Kwargs to pass onto the ``keras.layers.Layer`` superclass.
    """

    def __init__(self, num_components: int, dtype: tf.DType = tf.int32, **kwargs):
        super().__init__(num_components, dtype=dtype, **kwargs)

    def _build_distribution(self, shape: Tuple[Optional[int], ...]) -> None:
        self._indicator = _Indicator(self.num_components, dtype=self.dtype)

    def _get_distribution(self) -> distributions.Distribution:
        return self._indicator


class _Indicator(distributions.Distribution):
    def __init__(
        self, num_components: int, dtype: tf.DType = tf.int32, name: str = "Indicator"
    ):
        self._num_components = num_components
        super(_Indicator, self).__init__(
            dtype=dtype,
            name=name,
            reparameterization_type=NOT_REPARAMETERIZED,
            allow_nan_stats=True,
            validate_args=False,
        )

    def _log_prob(self, x: tf.Tensor) -> tf.Tensor:
        indicator_values = tf.one_hot(
            x, depth=self._num_components, on_value=0.0, off_value=float("-inf")
        )
        rank = len(indicator_values.shape)
        indicator_values = tf.transpose(
            indicator_values, list(range(rank - 2)) + [rank - 1, rank - 2]
        )
        return tf.squeeze(indicator_values, axis=3)



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow.keras import initializers


class Equidistant(initializers.Initializer):
    """
    Initializer that generates tensors where the last axis is initialized with 'equidistant' values.

    Args:
      minval: A python scalar or a scalar tensor. Lower bound of the range
        of random values to generate.
      maxval: A python scalar or a scalar tensor. Upper bound of the range
        of random values to generate.  Defaults to 1 for float types.
    """

    def __init__(self, minval: float = 0.0, maxval: float = 1.0):
        self.minval = minval
        self.maxval = maxval

    def __call__(
        self, shape: Tuple[Optional[int], ...], dtype: Optional[tf.DType] = None
    ):
        """
        Compute equidistant initializations on the last axis.

        Args:
            shape: Shape of Tensor to initialize.
            dtype: DType of Tensor to initialize.

        Returns:
            Initial value.

        Raises:
            ValueError: If shape cannot be determined.
        """
        rank = len(shape)
        last_dim = shape[-1]
        if last_dim is None:
            raise ValueError("Cannot compute Equidistant with unknown last dimension")
        linspace = tf.reshape(
            tf.linspace(self.minval, self.maxval, num=last_dim),
            [1] * (rank - 1) + [last_dim],
        )
        return tf.cast(
            tf.tile(linspace, tf.concat([shape[:-1], [1]], axis=0)), dtype=dtype
        )

    def get_config(self) -> dict:
        """
        Obtain config.

        Returns:
            Key-value mapping of the configuration.
        """
        return {"minval": self.minval, "maxval": self.maxval}



from typing import Optional

import tensorflow as tf
from tensorflow.keras.constraints import Constraint


class Clip(Constraint):
    """
    Constraints the weights to be between ``min`` and ``max``.

    Args:
        min: Minimum clip value
        max: Maximum clip value
    """

    def __init__(self, min: float, max: Optional[float] = None):
        self.min = min
        self.max = max

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        """
        Clip input tensor to ``[min, max]``.

        Args:
            w: Weight Tensor.

        Returns:
            Clipped weight Tensor.
        """
        return tf.clip_by_value(w, clip_value_min=self.min, clip_value_max=self.max)

    def get_config(self) -> dict:
        """
        Obtain config.

        Returns:
            Key-value mapping with configuration of this constraint.
        """
        return dict(min=self.min, max=self.max)



from .dynamic_spn import DynamicSumProductNetwork
from .sequential_spn import SequentialSumProductNetwork
from .spn import SumProductNetwork

__all__ = [
    "DynamicSumProductNetwork",
    "SumProductNetwork",
    "SequentialSumProductNetwork",
]



import tensorflow as tf
from tensorflow import keras


class NegativeLogJoint(keras.losses.Loss):
    r"""
    Compute :math:`-\log(p(X,Y))`.

    Assumes that its input is :math:`\log(p(X|Y))` where Y is indexed on the second axis. This can
    be used for supervised generative learning with gradient-based optimizers or
    (hard) expectation maximization.
    """

    def call(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        """
        Compute generative loss for optimizing :math:`p(X,Y)`.

        Args:
            y_true: Sparse labels
            y_pred: Predicted logits

        Returns:
            A `Tensor` that describes  a generative supervised loss.
        """
        return -tf.gather(y_pred, tf.cast(y_true, tf.int32), axis=1)



from typing import Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras import initializers


class PoonDomingosMeanOfQuantileSplit(initializers.Initializer):
    """
    Initializes the data according to the algorithm described in (Poon and Domingos, 2011).

    The data is divided over :math:`K` quantiles where :math:`K` is the number of nodes along
    the last axis of the tensor to be initialized. The quantiles are computed over all samples
    in the provided ``data``. Then, the mean per quantile is taken as the value for
    initialization.

    Args:
        data (numpy.ndarray): Data to compute quantiles over
        samplewise_normalization: Whether to 'Z-score normalize' the data sample-wise before
            computing the quantiles and means.
        normalization_epsilon: Non-zero constant to account for near-zero standard deviations in
            normalizations.

    References:
        Sum-Product Networks, a New Deep Architecture
        `Poon and Domingos, 2011 <https://arxiv.org/abs/1202.3732>`_
    """

    def __init__(
        self,
        data: Optional[np.ndarray] = None,
        samplewise_normalization: bool = True,
        normalization_epsilon: float = 1e-2,
    ):
        self._data = data
        self.samplewise_normalization = samplewise_normalization
        self.normalization_epsilon = normalization_epsilon

    def __call__(
        self, shape: Tuple[Optional[int], ...], dtype: tf.dtypes.DType = None
    ) -> tf.Tensor:
        """
        Compute initializations along the last axis.

        Args:
            shape: Shape of Tensor to initialize.
            dtype: DType of Tensor to initialize.

        Returns:
            Initial value.

        Raises:
            ValueError: If data was not provided first.
        """
        if self._data is None:
            raise ValueError(
                "Data was not set so cannot determine statistics for Poon & Domingos initialization"
            )

        num_quantiles = shape[-2]

        if self.samplewise_normalization:
            axes = tuple(range(1, len(self._data.shape)))
            data = (self._data - np.mean(self._data, axis=axes, keepdims=True)) / (
                np.std(self._data, axis=axes, keepdims=True)
                + self.normalization_epsilon
            )
        else:
            data = self._data

        batch_size = data.shape[0]
        quantile_sections = np.arange(
            batch_size // num_quantiles,
            batch_size,
            int(np.ceil(batch_size / num_quantiles)),
        )
        sorted_features = np.sort(data, axis=0)
        values_per_quantile = np.split(
            sorted_features, indices_or_sections=quantile_sections, axis=0
        )
        means_per_quantile = [
            np.mean(v, axis=0, keepdims=True) for v in values_per_quantile
        ]
        return tf.expand_dims(
            tf.cast(np.stack(means_per_quantile, axis=-1), dtype=dtype), axis=-1
        )

    def get_config(self) -> dict:
        """
        Obtain config.

        Returns:
            Key-value mapping of the configuration.
        """
        return {
            "samplewise_normalization": self.samplewise_normalization,
            "normalization_epsilon": self.normalization_epsilon,
        }



import tensorflow as tf

from libspn_keras.math.logutils import replace_infs_with_zeros


def logmatmul(log_a: tf.Tensor, log_b: tf.Tensor) -> tf.Tensor:
    """
    Matrix multiplication in log-space.

    Args:
        log_a: log(a) of shape [..., batch, num_in]
        log_b: log(b) of shape [..., num_in, num_out]

    Returns:
        A matrix log(c) where log(c) = log(a @ b)
    """
    with tf.name_scope("LogMatmul"):
        # Compute max for each tensor for numerical stability
        max_a = replace_infs_with_zeros(
            tf.stop_gradient(tf.reduce_max(log_a, axis=-1, keepdims=True))
        )
        max_b = replace_infs_with_zeros(
            tf.stop_gradient(tf.reduce_max(log_b, axis=-2, keepdims=True))
        )

        # Compute logsumexp using matrix multiplication
        return (
            tf.math.log(tf.matmul(tf.exp(log_a - max_a), tf.exp(log_b - max_b)))
            + max_a
            + max_b
        )



from typing import Optional, Tuple

import tensorflow as tf

from libspn_keras.constraints import GreaterEqualEpsilonNormalized, LogNormalized
from libspn_keras.layers.dense_sum import DenseSum
from libspn_keras.logspace import logspace_wrapper_initializer


class Conv2DSum(DenseSum):
    """
    Computes a convolutional sum, i.e. weights are shared across the spatial axes.

    Args:
        num_sums: Number of sums per spatial cell. Corresponds to the number of channels in
            the output
        logspace_accumulators: If ``True``, accumulators will be represented in log-space which
            is typically used with ``BackpropMode.GRADIENT``. If ``False``, accumulators will be
            represented in linear space. Weights are computed by normalizing the accumulators
            per sum, so that we always end up with a normalized SPN. If ``None`` (default) it
            will be set to ``True`` for ``BackpropMode.GRADIENT`` and ``False`` otherwise.
        accumulator_initializer: Initializer for accumulator
        sum_op (SumOpBase): SumOpBase instance which determines how to compute the forward and
            backward pass of the weighted sums
        accumulator_regularizer: Regularizer for accumulators
        linear_accumulator_constraint: Constraint for accumulators (only applied if
            log_space_accumulators==False)
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this leaf layer.

        Args:
            input_shape: Shape of the input Tensor.
        """
        # Create a trainable weight variable for this layer.
        _, num_scopes_vertical, num_scopes_horizontal, num_channels_in = input_shape

        weights_shape = (1, 1, num_channels_in, self.num_sums)

        initializer = self.accumulator_initializer
        accumulator_constraint = self.linear_accumulator_constraint
        if self.logspace_accumulators:
            initializer = logspace_wrapper_initializer(initializer)
            accumulator_constraint = self.logspace_accumulator_constraint

        self._forward_normalize = not isinstance(
            accumulator_constraint, (GreaterEqualEpsilonNormalized, LogNormalized)
        )
        self._accumulators = self.add_weight(
            name="sum_weights",
            shape=weights_shape,
            initializer=initializer,
            regularizer=self.accumulator_regularizer,
            constraint=accumulator_constraint,
        )
        if accumulator_constraint is not None:
            self._accumulators.assign(accumulator_constraint(self._accumulators))
        super(DenseSum, self).build(input_shape)

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute a convolutional sum, using 1x1 convolutions.

        Args:
            x: Spatial Tensor.
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with the same spatial dimensions and a number of channels determined
            by the number of channels set at the layer's instantiation.
        """
        return self.sum_op.weighted_conv(
            x,
            accumulators=self._accumulators,
            logspace_accumulators=self.logspace_accumulators,
            normalize_in_forward_pass=self._forward_normalize,
        )



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras


class Undecompose(keras.layers.Layer):
    """
    Undecomposes the incoming tensor by aligning the decomposition axis on the final node axis.

    Can only be done if the number of scopes (at the very first dimension of the input) is 1.

    Args:
        **kwargs: kwargs to pass onto the keras.Layer super class
    """

    def __init__(self, num_decomps: int = 1, **kwargs):
        super(Undecompose, self).__init__(**kwargs)
        self.num_decomps = num_decomps

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: If shape cannot be determined.
        """
        _, self._num_scopes, num_decomps_in, nodes_in = input_shape

        if num_decomps_in is None:
            raise ValueError("Cannot build Undecompose with unknown number of decomps.")

        if nodes_in is None:
            raise ValueError(
                "Cannot build Undecompose with unknown number of input nodes."
            )

        if num_decomps_in % self.num_decomps != 0:
            raise ValueError(
                "Number of decomps in input must be multiple of target number of decomps, got "
                "{} for input decomps and {} for target decomps.".format(
                    num_decomps_in, self.num_decomps
                )
            )

        number_of_decomps_to_join = num_decomps_in // self.num_decomps
        self._num_nodes = number_of_decomps_to_join * nodes_in

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Undecompose scopes by aligning neighboring decompositions along the node axis.

        Args:
            x: Decomposed Region input tensor.
            kwargs: Remaining keyword arguments.

        Returns:
            Undecomposed tensor.
        """
        shape = [-1, self._num_scopes, self.num_decomps, self._num_nodes]
        return tf.reshape(x, shape)

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.

        Raises:
            ValueError: When shape cannot be determined.
        """
        num_batch, num_scopes_in, num_decomps_in, num_nodes_in = input_shape
        if num_decomps_in is None:
            raise ValueError(
                "Cannot compute output shape with unknown number of input decomps"
            )
        if num_nodes_in is None:
            raise ValueError(
                "Cannot compute output shape with unknown number of input nodes"
            )
        number_of_decomps_to_join = num_decomps_in // self.num_decomps
        return (
            num_batch,
            num_scopes_in,
            self.num_decomps,
            num_nodes_in * number_of_decomps_to_join,
        )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(num_decomps=self.num_decomps,)
        base_config = super(Undecompose, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Optional

from tensorflow.keras.constraints import Constraint

from libspn_keras.constraints.greater_equal_epsilon_normalized import (
    GreaterEqualEpsilonNormalized,
)


_DEFAULT_LINEAR_ACCUMULATOR_CONSTRAINT: Optional[
    Constraint
] = GreaterEqualEpsilonNormalized()


def get_default_linear_accumulators_constraint() -> Optional[Constraint]:
    """
    Get default linear accumulator constraint.

    Returns:
        A ``Constraint`` instance that was set with ``set_default_linear_accumulators_constraint``
    """
    return _DEFAULT_LINEAR_ACCUMULATOR_CONSTRAINT


def set_default_linear_accumulators_constraint(op: Constraint) -> None:
    """
    Set default sum op to conveniently use it throughout an SPN architecture.

    Args:
        op: A constraint applied to linear accumulators after updates
    """
    global _DEFAULT_LINEAR_ACCUMULATOR_CONSTRAINT
    _DEFAULT_LINEAR_ACCUMULATOR_CONSTRAINT = op



import tensorflow as tf
from tensorflow import keras


class NegativeLogLikelihood(keras.losses.Loss):
    r"""
    Marginalize logits over last dimension so that it computes :math:`-\log(p(X))`.

    This can be used for unsupervised generative learning.
    """

    def call(self, _: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        r"""
        Compute generative loss for optimizing :math:`\log(p(X))`.

        Args:
            _: True labels, ignored, but still provided to preserve Keras compatibility
            y_pred: Predicted logits (or already marginalized region_graph_root value)

        Returns:
            A `Tensor` that describes a generative unsupervised loss.
        """
        return -tf.reduce_logsumexp(y_pred, axis=1)



from libspn_keras.initializers.dirichlet import Dirichlet
from libspn_keras.initializers.epsilon_inverse_fan_in import EpsilonInverseFanIn
from libspn_keras.initializers.equidistant import Equidistant
from libspn_keras.initializers.kmeans import KMeans
from libspn_keras.initializers.poon_domingos import PoonDomingosMeanOfQuantileSplit

__all__ = [
    "Dirichlet",
    "EpsilonInverseFanIn",
    "Equidistant",
    "PoonDomingosMeanOfQuantileSplit",
    "KMeans",
]



import abc
from collections import deque, OrderedDict
import functools
import itertools
import operator
from typing import Callable, Dict, Iterator, List, Mapping, Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras.constraints import Constraint
from tensorflow.keras.initializers import Initializer

from libspn_keras.layers import BaseLeaf, DenseProduct, DenseSum, RootSum
from libspn_keras.layers.flat_to_regions import FlatToRegions
from libspn_keras.layers.permute_and_pad_scopes import PermuteAndPadScopes
from libspn_keras.sum_ops import SumOpGradBackprop


class OverlappingScopesException(Exception):
    """Thrown when scopes of children overlap."""


class Region(abc.ABC):
    """
    Represents a region in the SPN.

    Regions graphs are DAGs just like SPNs, but they do not represent sums and products. They are merely
    used for defining the scope structure of the SPN.
    """

    @property
    @abc.abstractmethod
    def scope(self):  # noqa: ANN201
        """Obtain scope as a list of nodes."""

    @property
    @abc.abstractmethod
    def children(self):  # noqa: ANN201
        """Obtain children as a list of nodes."""

    @staticmethod
    def _assert_no_scope_overlap(children) -> None:  # noqa: ANN001
        """
        Ensure validity of the SPN represented by asserting none of the children overlap in their scopes.

        Args:
            children: List of Regions.

        Raises:
            OverlappingScopesException: the scopes of two children overlap.
        """
        for c0, c1 in itertools.combinations(children, 2):
            if set(c0.scope) & set(c1.scope):
                raise OverlappingScopesException(
                    "Children {} and {} have overlapping scopes".format(c0, c1)
                )


class RegionNode(Region):
    """
    Represents a region in the SPN.

    Regions graphs are DAGs just like SPNs, but they do not represent sums and products. They are merely
    used for defining the scope structure of the SPN.

    A RegionNode is a node with children, as opposed to a RegionVariable, which is a leaf node.

    Args:
          children (list of RegionNode or RegionVariable): children of this node. Scope of the
            resulting node is the union of the scopes of its children. The scopes of these children
            must not overlap (pairwise disjoint)
    """

    def __init__(self, children: List[Region]):
        self._assert_no_scope_overlap(children)
        self._children = children
        self._scope = list(
            functools.reduce(operator.concat, [c.scope for c in children])  # type: ignore
        )

    def __repr__(self):
        return "{" + ", ".join(str(s) for s in self._children) + "}"

    @property
    def children(self) -> List[Region]:
        """
        Obtain children as a list of nodes.

        Returns:
            A list of Regions
        """
        return self._children

    @property
    def scope(self) -> List[Region]:
        """
        Obtain scope as a list of nodes.

        Returns:
            A list of Regions
        """
        return self._scope


class RegionVariable(Region):
    """
    Represents a region in the SPN.

    Regions graphs are DAGs just like SPNs, but they do not represent sums and products. They are merely
    used for defining the scope structure of the SPN.

    A RegionVariable is a leaf node without children, as opposed to a RegionNode which is a node with
    children.

    Args:
        index: Index of the variable
    """

    def __init__(self, index: int):
        self.index = index

    def __repr__(self):
        return "x{}".format(self.index)

    @property
    def children(self) -> List[Region]:
        """
        Obtain children as a list of nodes.

        Returns:
            A list of Regions
        """
        return []

    @property
    def scope(self) -> List[Region]:
        """
        Obtain scope as a list of nodes.

        Returns:
            A list of Regions
        """
        return [self]


def region_graph_to_dense_spn(
    region_graph_root: RegionNode,
    leaf_node: BaseLeaf,
    num_sums_iterable: Iterator[int],
    return_weighted_child_logits: bool,
    logspace_accumulators: bool = False,
    accumulator_initializer: Optional[Initializer] = None,
    linear_accumulator_constraint: Optional[Constraint] = None,
    product_first: bool = True,
    num_classes: Optional[int] = None,
    with_root: bool = True,
) -> tf.keras.Sequential:
    """
    Convert a region graph (built from :class:`RegionNode` and :class:`RegionVar`) to a dense SPN.

    Args:
        region_graph_root: Root of the region graph
        leaf_node: Node to insert at the leaf of the SPN
        num_sums_iterable: Number of sums for all but the last root sum
            layer from bottom to top
        logspace_accumulators: Whether to represent accumulators of weights in logspace
            or not
        accumulator_initializer: Initializer for accumulators
        linear_accumulator_constraint: Constraint for linear
            accumulator, default: GreaterThanEpsilon
        product_first: Whether to start with a product layer
        num_classes: Number of classes at output. If ``None``, will not use 'latent' sums
            at the end but will instead directly connect the root to the final layer of the dense
            stack. This means if set to ``None`` the SPN cannot be used for classification.
        with_root: If ``True``, sets a ``RootSum`` as the final layer.
        return_weighted_child_logits: Whether to return weighted child logits. If ``

    Returns:
        A Sum-Product Network as a tf.keras.Sequential model.
    """
    (
        permutation,
        num_factors_leaf_to_root,
    ) = _region_graph_to_permutations_and_prods_per_depth(region_graph_root)

    sum_kwargs = dict(
        logspace_accumulators=logspace_accumulators,
        accumulator_initializer=accumulator_initializer,
        linear_accumulator_constraint=linear_accumulator_constraint,
        sum_op=SumOpGradBackprop(),
    )

    sum_product_stack: List[tf.keras.layers.Layer] = []
    if not product_first:
        sum_product_stack.append(
            DenseSum(num_sums=next(num_sums_iterable), **sum_kwargs)
        )
    for depth, num_factors in enumerate(num_factors_leaf_to_root):
        sum_product_stack.append(DenseProduct(num_factors=num_factors))
        if depth == len(num_factors_leaf_to_root) - 1:
            break
        sum_product_stack.append(
            DenseSum(num_sums=next(num_sums_iterable), **sum_kwargs)
        )

    if num_classes is not None:
        sum_product_stack.append(DenseSum(num_sums=num_classes, **sum_kwargs))  # type: ignore

    if with_root:
        sum_product_stack.append(
            RootSum(
                return_weighted_child_logits=return_weighted_child_logits, **sum_kwargs  # type: ignore
            )
        )

    pre_stack: List[tf.keras.layers.Layer] = [
        FlatToRegions(
            num_decomps=1, input_shape=[len(_collect_variable_nodes(region_graph_root))]
        ),
        leaf_node,
        PermuteAndPadScopes(permutations=np.asarray([permutation])),
    ]

    return tf.keras.Sequential(pre_stack + sum_product_stack)


def _get_nodes_to_depth_mapping(root: Region) -> Dict[Region, int]:
    node_to_depth = dict()
    node_to_depth[root] = 0

    def _node_fn(node: Region) -> None:
        # Add to Parents dict
        for child in node.children:
            node_to_depth[child] = node_to_depth[node] + 1

    _traverse_region_graph(root, _node_fn)
    return node_to_depth


def _get_depth_to_nodes_mapping(
    root: Region, node_to_depth_mapping: Optional[Mapping[Region, int]] = None,
) -> Dict[int, List[Region]]:

    node_to_depth_mapping = node_to_depth_mapping or _get_nodes_to_depth_mapping(root)
    depth_to_nodes = OrderedDict()

    for node, depth in sorted(
        node_to_depth_mapping.items(), key=operator.itemgetter(1)
    ):
        if depth not in depth_to_nodes:
            depth_to_nodes[depth] = [node]
        else:
            depth_to_nodes[depth].append(node)

    return depth_to_nodes


def _compute_max_num_children_by_depth(
    root: Region, depth_to_nodes_mapping: Mapping[int, List[Region]] = None,
) -> Dict[int, int]:
    depth_to_nodes_mapping = depth_to_nodes_mapping or _get_depth_to_nodes_mapping(root)
    max_num_children_by_depth = OrderedDict()
    for depth, nodes in depth_to_nodes_mapping.items():
        if any(isinstance(n, RegionNode) for n in nodes):
            max_num_children_by_depth[depth] = max(
                len(node.children) for node in nodes if isinstance(node, RegionNode)
            )
    return max_num_children_by_depth


def _collect_variable_nodes(root: RegionNode) -> List[Region]:
    variable_nodes: List[Region] = []

    def _gather(node: Region) -> None:
        if isinstance(node, RegionVariable):
            variable_nodes.append(node)
        else:
            for child in node.children:
                _gather(child)

    _gather(root)
    return variable_nodes


def _region_graph_to_permutations_and_prods_per_depth(
    root: RegionNode,
) -> Tuple[list, List[int]]:
    node_to_depth_mapping = _get_nodes_to_depth_mapping(root)
    depth_to_nodes_mapping = _get_depth_to_nodes_mapping(
        root, node_to_depth_mapping=node_to_depth_mapping
    )
    max_num_children_by_depth = _compute_max_num_children_by_depth(
        root, depth_to_nodes_mapping=depth_to_nodes_mapping
    )
    max_num_children_by_depth_cum_prod = np.cumprod(
        [1] + list(reversed(list(max_num_children_by_depth.values())))
    )
    max_depth = max(node_to_depth_mapping.values())

    permutations = []

    def recurse_permutation(node: Region) -> None:
        depth = node_to_depth_mapping[node]
        if isinstance(node, RegionVariable) and depth < max_depth:
            permutations.append(node.index)
            for _ in range(
                max_num_children_by_depth_cum_prod[max_depth]
                // max_num_children_by_depth_cum_prod[depth + 1]
            ):
                permutations.append(-1)
        elif isinstance(node, RegionVariable) and depth == max_depth:
            permutations.append(node.index)
        elif isinstance(node, RegionNode):
            for child in node.children:
                recurse_permutation(child)
            for _ in range(
                max_num_children_by_depth_cum_prod[depth]
                // max_num_children_by_depth_cum_prod[depth + 1]
                - len(node.children)
            ):
                children_beneath = (
                    max_num_children_by_depth_cum_prod[max_depth]
                    // max_num_children_by_depth_cum_prod[depth + 1]
                )
                permutations.extend([-1] * children_beneath)

    recurse_permutation(root)

    return permutations, list(reversed(list(max_num_children_by_depth.values())))


def _traverse_region_graph(root: Region, fun: Callable[[Region], None]) -> None:
    """
    Run ``fun`` on descendants of ``root`` (including ``root``) breadth-first.

    Args:
        root (Region): The root of the SPN graph.
        fun (function): A function ``fun(node)`` executed once for every node of
                        the graph. It should return ``True`` if traversing
                        should be stopped.
    """
    visited_nodes = set()  # Set of visited nodes
    queue: deque = deque()
    queue.append(root)

    while queue:
        next_node = queue.popleft()
        if next_node not in visited_nodes:
            fun(next_node)

            visited_nodes.add(next_node)

            for child in next_node.children:
                queue.append(child)



from typing import List, Optional, Tuple

import tensorflow as tf
from tensorflow import keras


class TemporalDenseProduct(keras.layers.Layer):
    r"""
    Computes 'temporal' dense products.

    This is used to connect an interface stack at :math:`t - 1` of a dynamic
    SPN with a template SPN at :math:`t`. Computes a product of all possible combinations of
    nodes along the last axis of the two incoming layers.

    Args:
        **kwargs: kwargs to pass on to the keras.Layer super class.
    """

    def __init__(self, **kwargs):
        self._num_nodes_out = None
        super(TemporalDenseProduct, self).__init__(**kwargs)

    def build(self, input_shape: List[Tuple[Optional[int], ...]]) -> None:
        """
        Build the internal components for this leaf layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: If shape cannot be determined.
        """
        a_shape, b_shape = input_shape
        if a_shape[-1] is None:
            raise ValueError("Unknown number of nodes in first tensor")
        if b_shape[-1] is None:
            raise ValueError("Unknown number of nodes in second tensor")
        self._num_nodes_out = a_shape[-1] * b_shape[-1]
        super(TemporalDenseProduct, self).build(input_shape)

    def call(self, x: List[tf.Tensor], **kwargs) -> tf.Tensor:
        """
        Compute temporal dense product.

        Multiplies every node at timestep t-1 with every node at timestep t, across all scopes
        and decompositions.

        Args:
            x: Spatial or Region tensor
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with a dense temporal product.
        """
        # Split in list of tensors which will be added up using outer products
        a, b = x
        a_expanded = tf.expand_dims(a, axis=-1)
        b_expanded = tf.expand_dims(b, axis=-2)
        outer_dims = tf.shape(a_expanded)[:-2]
        out_shape = tf.concat([outer_dims, [self._num_nodes_out]], axis=0)
        return tf.reshape(a_expanded + b_expanded, out_shape)

    def compute_output_shape(
        self, input_shape: List[Tuple[Optional[int], ...]]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.

        Raises:
            ValueError: If shape cannot be determined.
        """
        a_shape, b_shape = input_shape
        num_batch, num_scopes_in, num_decomps, num_nodes_in_a = a_shape

        if num_nodes_in_a is None:
            raise ValueError(
                "Cannot compute output shape without knowing num nodes in first input"
            )

        if b_shape[-1] is None:
            raise ValueError(
                "Cannot compute output shape without knowing num nodes in second input"
            )

        return num_batch, num_scopes_in, num_decomps, int(num_nodes_in_a ** b_shape[-1])



import abc
from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras
import tensorflow_probability as tfp


class BaseLeaf(keras.layers.Layer, abc.ABC):
    """
    Computes probabilities from raw input.

    Args:
        num_components: Number of components per variable.
        dtype: DType of the input.
    """

    def __init__(self, num_components: int, dtype: tf.DType = tf.float32, **kwargs):
        super(BaseLeaf, self).__init__(dtype=dtype, **kwargs)
        self.num_components = num_components

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this leaf layer.

        Args:
            input_shape: Shape of the input Tensor.
        """
        _, *scope_dims, multivariate_size = input_shape
        distribution_shape = 1, *scope_dims, self.num_components, multivariate_size
        self._num_scopes, self._num_decomps = scope_dims
        self._build_distribution(distribution_shape)
        super(BaseLeaf, self).build(input_shape)

    @abc.abstractmethod
    def _build_distribution(self, shape: Tuple[Optional[int], ...]) -> None:
        pass

    @abc.abstractmethod
    def _get_distribution(self) -> tfp.distributions.Distribution:
        pass

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute the probability of the leaf nodes.

        Args:
            x: Spatial or region Tensor with raw input values.
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with the probabilities per component.
        """
        x = tf.expand_dims(x, axis=-2)
        distribution = self._get_distribution()
        return tf.reduce_sum(distribution.log_prob(x), axis=-1)

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.
        """
        *outer_dims, _ = input_shape
        return (*outer_dims, self.num_components)

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(num_components=self.num_components)
        base_config = super(BaseLeaf, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def get_modes(self) -> tf.Tensor:
        """
        Obtain the distribution modes.

        This can be used for e.g. MPE estimates of inputs.

        Raises:
            NotImplementedError: Not all descendants of BaseLeaf have this implemented.Not
        """
        raise NotImplementedError(
            "A {} does not implement distribution modes.".format(
                self.__class__.__name__
            )
        )



from collections import OrderedDict
import functools
import itertools
import operator
from typing import Any, List

import colorlover as cl
import numpy as np
import plotly.graph_objects as go
import tensorflow as tf

from libspn_keras.layers import (
    DenseProduct,
    DenseSum,
    FlatToRegions,
    PermuteAndPadScopes,
    RootSum,
)


def visualize_dense_spn(
    dense_spn: tf.keras.Model,
    show_legend: bool = False,
    show_padding: bool = True,
    transparent: bool = False,
    node_size: int = 30,
) -> go.Figure:
    """
    Visualize dense SPN, consisting of ``DenseSum``, ``DenseProduct``, ``RootSum`` and leaf layers.

    Args:
        dense_spn: An SPN of type ``tensorflow.keras.Sequential``
        show_legend: Whether to show legend of scopes and layers on the right
        show_padding: Whether to show padded nodes
        transparent: If ``True``, the background is transparent.
        node_size: Size of the nodes drawn in the graph. Adjust to avoid clutter.

    Returns:
        A ``plotly.graph_objects.Figure`` instance. Use ``.show()`` to render the visualization.
    """
    (
        nodes,
        edges,
        colors,
        symbols,
        scopes,
        node_group_sizes,
        names,
    ) = _assemble_dense_spn_figure(dense_spn, show_padding=show_padding)

    hovertext = [
        "{" + ", ".join(str(s) for s in sorted(scope)) + "}"
        for scope in scopes.values()
    ]

    y_n, x_n = zip(*nodes.values())

    # Build figure
    fig = go.Figure(
        layout=go.Layout(
            plot_bgcolor="rgb(255,255,255,0)",
            xaxis=go.layout.XAxis(ticks="", tickvals=[]),
            yaxis=go.layout.YAxis(ticks="", tickvals=[]),
            legend=dict(traceorder="reversed", font=dict(size=18)),
        )
    )
    if transparent:
        fig.layout.paper_bgcolor = "rgb(255,255,255,0)"

    y_e, x_e = zip(*np.asarray(edges).transpose((0, 2, 1)))

    x_e_flat = list(x_e[0]) + functools.reduce(
        operator.concat, [[None] + list(x_ei) for x_ei in x_e]  # type: ignore
    )
    y_e_flat = list(y_e[0]) + functools.reduce(
        operator.concat, [[None] + list(y_ei) for y_ei in y_e]  # type: ignore
    )

    fig.add_trace(
        go.Scatter(
            mode="lines",
            x=x_e_flat,
            y=y_e_flat,
            showlegend=False,
            line=dict(color="gray"),
        )
    )

    # Add scatter trace with medium sized markers
    offset = 0
    for group_size, name in zip(node_group_sizes, names):
        group_ind = slice(offset, offset + group_size)
        offset += group_size
        if not show_padding and all(s == "asterisk" for s in symbols[group_ind]):
            continue
        fig.add_trace(
            go.Scatter(
                mode="markers",
                x=x_n[group_ind],
                y=y_n[group_ind],
                hovertext=hovertext[group_ind],
                hoverinfo="text",
                marker=dict(
                    color=colors[group_ind],
                    size=node_size,
                    symbol=symbols[group_ind],
                    line=dict(width=2),
                ),
                showlegend=show_legend,
                name=name,
            )
        )
    return fig


def _assemble_dense_spn_figure(  # noqa: C901
    dense_spn: tf.keras.Model, show_padding: bool = True
):  # noqa: C901, ANN202
    color_palette = itertools.cycle(cl.scales["12"]["qual"]["Set3"])
    max_total_nodes = 0
    for layer in dense_spn.layers:
        if isinstance(layer, RootSum):
            num_scopes = 1
            num_nodes = 1
        else:
            _, num_scopes, _, num_nodes = layer.output_shape
        total_nodes = num_scopes * num_nodes
        max_total_nodes = max(total_nodes, max_total_nodes)

    node_yx = OrderedDict()
    edges_yx: List[Any] = []

    colors = []
    symbols = []
    scopes = OrderedDict()

    last_width = 0.0
    last_num_nodes = None
    prev_layer = None
    node_group_sizes = []
    names = []
    for layer_index, layer in enumerate(dense_spn.layers):

        if isinstance(layer, FlatToRegions):
            continue

        if isinstance(layer, RootSum):
            num_scopes = 1
            num_nodes = 1
        else:
            _, num_scopes, _, num_nodes = layer.output_shape
        current_width = float(num_scopes * num_nodes)
        if current_width < last_width:
            if last_num_nodes == num_nodes * num_scopes:
                current_width = last_width
            else:
                current_width = (last_width + current_width) / 2 + 1
        xs = (
            np.linspace(
                -current_width / 2 + 0.5,
                current_width / 2 - 0.5,
                num=num_scopes * num_nodes,
            )
            if num_scopes * num_nodes != 1
            else [0]
        )
        for scope_index in range(num_scopes):
            ci = next(color_palette)
            for nj, xj in enumerate(
                xs[num_nodes * scope_index : num_nodes * (scope_index + 1)]
            ):
                node_yx[(layer_index, scope_index, nj)] = (layer_index, xj)
                colors.append(ci)
                if isinstance(layer, (DenseSum, RootSum)):
                    symbols.append("circle-cross")
                elif isinstance(layer, DenseProduct):
                    symbols.append("circle-x")
                elif isinstance(layer, PermuteAndPadScopes):
                    if layer.permutations[0][scope_index] == -1:  # type: ignore
                        symbols.append("asterisk")
                    else:
                        symbols.append("circle")
                else:
                    symbols.append("circle")

                if prev_layer is not None:
                    scope = set()
                    num_nodes_prev = prev_layer.output_shape[-1]
                    if isinstance(layer, (DenseSum, RootSum)):
                        edges_yx.extend(
                            [
                                (
                                    (layer_index, xj),
                                    node_yx[
                                        (layer_index - 1, scope_index, node_offset)
                                    ],
                                )
                                for node_offset in range(num_nodes_prev)
                            ]
                        )
                        scope = scopes[(layer_index - 1, scope_index, 0)]
                    elif isinstance(layer, DenseProduct):
                        nk = nj
                        for factor in range(layer.num_factors):
                            child_key = (
                                layer_index - 1,
                                scope_index * layer.num_factors + factor,
                                nk % num_nodes_prev,
                            )

                            if scopes[child_key] != set() or show_padding:
                                edges_yx.append(((layer_index, xj), node_yx[child_key]))
                            scope |= scopes[child_key]

                            nk //= num_nodes_prev

                    elif isinstance(layer, PermuteAndPadScopes):
                        prev_scope_index = layer.permutations[0, scope_index]
                        if prev_scope_index != -1:
                            child_key = (
                                layer_index - 1,
                                prev_scope_index,
                                nj % num_nodes,
                            )
                            edges_yx.append(((layer_index, xj), node_yx[child_key]))
                            scope = {int(prev_scope_index)}

                        else:
                            scope = set()
                else:
                    scope = {scope_index}

                scopes[(layer_index, scope_index, nj)] = scope

            node_group_sizes.append(num_nodes)

            if scope == set():
                scope_suffix = r"\emptyset"
            else:
                scope_suffix = (
                    "{" + ", ".join("X_{}".format(s) for s in sorted(scope)) + "}"
                )
            names.append(
                "$\\text{" + layer.__class__.__name__ + " }" + scope_suffix + "$"
            )

        prev_layer = layer
        last_width = current_width
        last_num_nodes = num_scopes * num_nodes

    return node_yx, edges_yx, colors, symbols, scopes, node_group_sizes, names



import tensorflow as tf


def replace_infs_with_zeros(x: tf.Tensor) -> tf.Tensor:
    """
    Replace infinite values with zeroes.

    Args:
        x: A `Tensor`.

    Returns:
        A `Tensor` with the same shape an dtype as `x` where infinite values are replaced
        with zeroes.
    """
    with tf.name_scope("ReplaceInfsWithZeroes"):
        return tf.where(tf.math.is_inf(x), tf.zeros_like(x), x)



import itertools
from typing import Any, List, Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras import initializers


class KMeans(initializers.Initializer):
    """
    Initializer learned through K-means from data.

    The centroids learned from K-means are used to initialize the location parameters of a location-scale
    leaf, such as a ``NormalLeaf``. This is particularly useful for variables with dimensionality of
    greater than 1.


    Notes:
        Currently only works for spatial SPNs.

    Args:
        data (numpy.ndarray): Data on which to perform K-means.
        samplewise_normalization (bool): Whether to normalize data before learning centroids.
        data_fraction (float): Fraction of the data to use for K-means (chosen randomly)
        normalization_epsilon (float): Normalization constant (only used when
            ``sample_normalization`` is ``True``.
        stop_epsilon: Non-zero constant for difference in MSE on which to stop K-means fitting.
        num_iters (int): Maximum number of iterations.
        group_centroids (bool): If ``True``, performs another round of K-means to group the
            centroids along the scope axes.
        max_num_clusters (int): Maximum number of clusters (use this to limit the memory needed)
        jitter_factor (float): If the number of clusters is larger than allowed according to
            ``max_num_clusters``, the learned ``max_num_clusters`` centroids are repeated and
            then jittered with noise generated from a truncated normal distribution with a
            standard deviation of ``jitter_factor``
        centroid_initialization (str): Centroid initialization algorithm. If ``"kmeans++"``, will
            iteratively initialize clusters far apart from each other. Otherwise, the centroids
            will be initialized from the data randomly.
    """

    def __init__(
        self,
        data: Optional[np.ndarray] = None,
        samplewise_normalization: bool = True,
        data_fraction: float = 0.2,
        normalization_epsilon: float = 1e-2,
        stop_epsilon: float = 1e-4,
        num_iters: int = 100,
        group_centroids: bool = True,
        max_num_clusters: int = 8,
        jitter_factor: float = 0.05,
        centroid_initialization: str = "kmeans++",
        downsample: Optional[int] = None,
        use_groups: bool = False,
    ):
        self._data = data
        self.samplewise_normalization = samplewise_normalization
        self.normalization_epsilon = normalization_epsilon
        self.data_fraction = data_fraction
        self.stop_epsilon = stop_epsilon
        self.group_centroids = group_centroids
        self.num_iters = num_iters
        self.max_num_clusters = max_num_clusters
        self.jitter_factor = jitter_factor
        self.centroid_initialization = centroid_initialization
        self.downsample = downsample
        self.use_groups = use_groups

    def __call__(  # noqa: C901
        self, shape: Tuple[Optional[int], ...], dtype: Optional[tf.dtypes.DType] = None
    ) -> tf.Tensor:
        """
        Compute KMeans initializations on the last axis.

        Args:
            shape: Shape of Tensor to initialize.
            dtype: DType of Tensor to initialize.

        Returns:
            Initial value.

        Raises:
            ValueError: If shape cannot be determined.
        """
        data = self._data
        if data is None:
            raise ValueError(
                "Cannot compute KMeans initialization without provided data"
            )
        height, width = shape[1:3]
        if height is None:
            raise ValueError("Unknown height")
        if width is None:
            raise ValueError("Unknown width")
        if self.downsample is not None:
            data = tf.image.resize(
                data, size=(height // self.downsample, width // self.downsample)
            ).numpy()

        num_components = shape[-2]
        if num_components is None:
            raise ValueError("Unknown number of components")
        if (
            num_components > self.max_num_clusters
            and num_components % self.max_num_clusters != 0
        ):
            raise ValueError(
                "Number of components must be multiple of max number of clusters"
            )

        if self.samplewise_normalization:
            axes = tuple(range(1, len(data.shape)))
            data = (data - tf.reduce_mean(data, axis=axes, keepdims=True)) / (
                tf.math.reduce_std(data, axis=axes, keepdims=True)
                + self.normalization_epsilon
            )
            data = data.numpy()

        batch_size, *middle_dims, dimensionality = data.shape

        fraction_size = int(len(data) * self.data_fraction)
        indices = np.random.choice(np.arange(len(data)), size=fraction_size)

        data_by_kmeans_problem = (
            data[indices]
            .reshape([fraction_size, -1, dimensionality])
            .transpose((1, 0, 2))
            .astype(np.float32)
        )

        num_clusters = min(self.max_num_clusters, num_components)
        centroids = self._kmeans_tf(data_by_kmeans_problem, num_clusters=num_clusters)

        if self.group_centroids:
            centroids = self._group_centroids(centroids, num_clusters)
        if num_clusters < num_components:
            centroids = tf.expand_dims(centroids, axis=-2)
            noise = tf.random.normal(
                tf.concat(
                    [
                        centroids.shape[:-2],
                        [num_components // num_clusters],
                        centroids.shape[-1:],
                    ],
                    axis=0,
                ),
                stddev=self.jitter_factor,
            )
            centroids = centroids + noise

        if self.downsample:
            centroids = tf.reshape(
                centroids,
                (
                    1,
                    height // self.downsample,
                    width // self.downsample,
                    num_components,
                    centroids.shape[-1],
                ),
            )
            dims_prefix, last_two_dims = (
                tf.shape(centroids)[:-2],
                tf.shape(centroids)[-2:],
            )
            channels_and_centroids_on_last_axis = tf.reshape(
                centroids,
                tf.concat([dims_prefix, [last_two_dims[0] * last_two_dims[1]]], axis=0),
            )
            centroids = tf.reshape(
                tf.image.resize(channels_and_centroids_on_last_axis, (height, width)),
                [dims_prefix[0], height, width, last_two_dims[0], last_two_dims[1]],
            )

        return np.reshape(np.asarray(centroids), shape)

    def get_config(self) -> dict:
        """
        Obtain config.

        Returns:
            Key-value mapping of the configuration.
        """
        return {
            "samplewise_normalization": self.samplewise_normalization,
            "normalization_epsilon": self.normalization_epsilon,
            "num_iters": self.num_iters,
            "stop_epsilon": self.stop_epsilon,
            "group_centroids": self.group_centroids,
        }

    def _kmeans_tf(self, data: np.ndarray, num_clusters: int) -> tf.Tensor:

        num_problems, num_batch, num_dims = data.shape
        if self.centroid_initialization == "kmeans++":
            indices = tf.random.categorical(
                logits=tf.zeros([1, num_batch]), num_samples=num_problems
            )
            centroids = tf.gather(
                data, tf.transpose(indices, (1, 0)), axis=1, batch_dims=1
            )

            for _ in tf.range(num_clusters - 1):
                distances = tf.reduce_sum(
                    tf.math.squared_difference(
                        tf.expand_dims(centroids, axis=2), tf.expand_dims(data, axis=1)
                    ),
                    axis=-1,
                )
                min_distances = tf.reduce_min(distances, axis=1)
                logits = tf.math.log(min_distances)
                indices = tf.random.categorical(logits=logits, num_samples=1)
                new_centroids = tf.gather(data, indices, axis=1, batch_dims=1)
                centroids = tf.concat([centroids, new_centroids], axis=1)
        else:
            indices = tf.random.categorical(
                logits=tf.zeros([1, num_batch]), num_samples=num_clusters * num_problems
            )
            indices = tf.reshape(indices, (num_problems, num_clusters))
            centroids = tf.gather(data, indices, axis=1, batch_dims=1)
            centroids += tf.random.normal(
                centroids.shape, stddev=0.05, dtype=centroids.dtype
            )

        mse = None
        for _ in range(self.num_iters):
            centroids, mse_new = self._kmeans_step(data, centroids, num_clusters)
            if mse is not None and tf.abs(mse - mse_new) < self.stop_epsilon:
                break
            mse = mse_new
        return centroids

    def _group_centroids(self, centroids: tf.Tensor, num_clusters: int) -> tf.Tensor:
        flat_centroids = tf.reshape(centroids, (-1, centroids.shape[-1]))

        if self.centroid_initialization == "kmeans++":
            indices = tf.reshape(
                tf.random.categorical(
                    logits=tf.zeros([1, flat_centroids.shape[0]]), num_samples=1
                ),
                (-1,),
            )
            super_centroids = tf.gather(flat_centroids, indices, axis=0)
            for _ in tf.range(num_clusters - 1):
                distances = tf.reduce_sum(
                    tf.math.squared_difference(
                        tf.expand_dims(super_centroids, axis=1),
                        tf.expand_dims(flat_centroids, axis=0),
                    ),
                    axis=-1,
                )
                min_distances = tf.reduce_min(distances, axis=0, keepdims=True)
                logits = tf.math.log(min_distances)
                indices = tf.random.categorical(logits=logits, num_samples=1)
                new_super_centroids = tf.gather(
                    flat_centroids, tf.reshape(indices, (-1,)), axis=0
                )
                super_centroids = tf.concat(
                    [super_centroids, new_super_centroids], axis=0
                )
        else:
            indices = tf.reshape(
                tf.random.categorical(
                    logits=tf.zeros([1, flat_centroids.shape[0]]),
                    num_samples=num_clusters,
                ),
                (-1,),
            )
            super_centroids = tf.gather(flat_centroids, indices, axis=0)

        def generate_permutations(n: int) -> List[Tuple[Any]]:
            return list(*itertools.permutations(range(n)))

        permutations = generate_permutations(num_clusters)

        mse = tf.reduce_mean(
            tf.reduce_sum(
                tf.math.squared_difference(centroids, super_centroids), axis=-1
            ),
        )
        for _ in range(self.num_iters):
            super_centroids, centroids, new_mse = self._assign_to_supercentroid(
                centroids, super_centroids, permutations
            )
            if mse is not None and tf.abs(new_mse - mse) < self.stop_epsilon:
                break
            mse = new_mse

        if self.use_groups:
            return tf.tile(
                tf.expand_dims(super_centroids, axis=0), [tf.shape(centroids)[0], 1, 1]
            )

        return centroids

    def _assign_to_supercentroid(
        self, centroids: tf.Tensor, super_centroids: tf.Tensor, permutations: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        num_clusters = centroids.shape[1]
        distances = tf.reduce_sum(
            tf.math.squared_difference(
                tf.expand_dims(tf.expand_dims(super_centroids, axis=0), axis=2),
                tf.expand_dims(centroids, axis=1),
            ),
            axis=-1,
        )
        distances_flat = tf.reshape(distances, (-1, num_clusters * num_clusters))
        indices = permutations + tf.range(num_clusters) * num_clusters
        assignment_distances = tf.reduce_mean(
            tf.gather(distances_flat, indices, axis=1), axis=-1
        )
        assignment_argmin = tf.argmin(assignment_distances, axis=-1)
        assignments = tf.gather(permutations, assignment_argmin, axis=0)
        centroids = tf.gather(centroids, assignments, axis=1, batch_dims=1)
        super_centroids_new = tf.reduce_mean(centroids, axis=0)
        mse = tf.reduce_mean(
            tf.reduce_sum(
                tf.math.squared_difference(centroids, super_centroids_new), axis=-1
            ),
        )
        return super_centroids_new, centroids, mse

    def _assign_to_centroid(
        self, data: tf.Tensor, centroids: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        distances = tf.reduce_sum(
            tf.math.squared_difference(
                tf.expand_dims(centroids, axis=1), tf.expand_dims(data, axis=2)
            ),
            axis=-1,
        )
        mse = tf.reduce_mean(tf.reduce_min(distances, axis=2))
        return tf.argmin(distances, axis=2), mse

    @tf.function
    def _kmeans_step(
        self, data: tf.Tensor, centroids: tf.Tensor, num_clusters: int
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        assignment, mse = self._assign_to_centroid(data, centroids)

        def _compute_mean(x: tf.Tensor) -> tf.Tensor:
            return tf.cast(
                tf.math.unsorted_segment_mean(x[0], x[1], num_segments=num_clusters),
                tf.float32,
            )

        new_centroids = tf.map_fn(_compute_mean, (data, assignment), dtype=tf.float32)
        return new_centroids, mse



from typing import Optional

import tensorflow as tf
from tensorflow.keras.constraints import Constraint


class LogNormalized(Constraint):
    """
    Normalizes log-space weights.

    Args:
        axis (int): Axis along whichto normalize
    """

    def __init__(self, axis: Optional[int] = -2):
        self.axis = axis

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        """
        Normalize weight in log-space.

        Args:
            w: Unnormalized weight Tensor.

        Returns:
            Log-normalized weight Tensor.
        """
        return tf.nn.log_softmax(w, axis=self.axis)

    def get_config(self) -> dict:
        """
        Obtain config.

        Returns:
            Key-value mapping with configuration of this constraint.
        """
        return dict(axis=self.axis)



from typing import List, Optional, Tuple, Union

import numpy
import tensorflow as tf
from tensorflow import keras


class PermuteAndPadScopes(keras.layers.Layer):
    """
    Permutes and pads scopes, usually applied after a ``FlatToRegions`` and a ``BaseLeaf`` layer.

    Padding can be used to ensure uniform dimension sizes across scopes, decomps and nodes in the
    layer stack. Padding is achieved by using ``-1`` for scope indices in ``permutations``.

    Args:
        num_decomps: Number of decompositions
        permutations: If None, permutations must be specified later
        **kwargs: kwargs to pass on to the keras.Layer superclass.
    """

    def __init__(
        self,
        permutations: Optional[Union[List[List[int]], numpy.ndarray]] = None,
        **kwargs
    ):
        super(PermuteAndPadScopes, self).__init__(**kwargs)
        self.permutations = permutations

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Permute and pad scopes.

        Padding is achieved by using ``-1`` for scope indices in ``permutations``/.

        Args:
            x: A Region input tensor
            kwargs: Remaining keyword arguments.

        Returns:
            Tensor with scopes permuted and padded.
        """
        decomps_first = tf.transpose(x, (2, 1, 0, 3))
        decomps_first_padded = tf.pad(decomps_first, [[0, 0], [1, 0], [0, 0], [0, 0]])
        gather_indices = tf.convert_to_tensor(self.permutations) + 1
        permuted = tf.gather(decomps_first_padded, gather_indices, axis=1, batch_dims=1)
        return tf.transpose(permuted, (2, 1, 0, 3))

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.
        """
        return input_shape

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(permutations=self.permutations)
        base_config = super(PermuteAndPadScopes, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import abc
from typing import Optional, Tuple, Union

import tensorflow as tf
from tensorflow import initializers
import tensorflow_probability as tfp
from tensorflow_probability.python.math import softplus_inverse

from libspn_keras.constraints import GreaterEqualEpsilon
from libspn_keras.layers.base_leaf import BaseLeaf
from libspn_keras.math.soft_em_grads import (
    LocationEMGradWrapper,
    LocationScaleEMGradWrapper,
)


class LocationScaleLeafBase(BaseLeaf, abc.ABC):
    """
    Computes the log probability of multiple components per variable along the final axis.

    Args:
        num_components: Number of components per variable
        location_initializer: Initializer for location variable
        location_trainable: Boolean that indicates whether location is trainable
        scale_initializer: Initializer for scale variable
        scale_constraint: Constraint for scale parameter
        **kwargs: kwar

        gs to pass on to the keras.Layer super class
    """

    def __init__(
        self,
        num_components: int,
        location_initializer: Optional[tf.keras.initializers.Initializer] = None,
        location_trainable: bool = True,
        scale_initializer: Optional[tf.keras.initializers.Initializer] = None,
        scale_trainable: bool = False,
        accumulator_initializer: Optional[tf.keras.initializers.Initializer] = None,
        use_accumulators: bool = False,
        scale_constraint: Optional[tf.keras.constraints.Constraint] = None,
        **kwargs
    ):
        super(LocationScaleLeafBase, self).__init__(
            num_components=num_components, **kwargs
        )
        self.location_initializer = (
            location_initializer or initializers.TruncatedNormal(stddev=1.0)
        )
        self.location_trainable = location_trainable
        self.scale_initializer = scale_initializer or initializers.Ones()
        self.scale_trainable = scale_trainable
        self.accumulator_initializer = accumulator_initializer or initializers.Ones()
        self.use_accumulators = use_accumulators
        self.scale_constraint = scale_constraint
        self._num_scopes = self._num_decomps = None

    def _build_distribution(self, shape: Tuple[Optional[int], ...]) -> None:
        if self.use_accumulators:
            self._create_loc_scale_accumulators(shape)
        else:
            self._create_loc_scale_vars(shape)

    def _get_distribution(
        self,
    ) -> Union[
        tfp.distributions.Distribution,
        LocationEMGradWrapper,
        LocationScaleEMGradWrapper,
    ]:
        if self.use_accumulators:
            return self._get_distribution_from_accumulators()
        return self._get_distribution_from_vars()

    def _get_distribution_from_vars(self) -> tfp.distributions.Distribution:
        if self.scale_trainable and not self.use_accumulators:
            scale = tf.nn.softplus(self.scale)
        else:
            scale = self.scale
        if self.scale_constraint is not None:
            scale = self.scale_constraint(scale)
        return self._build_distribution_from_loc_and_scale(loc=self.loc, scale=scale)

    def _get_distribution_from_accumulators(
        self,
    ) -> Union[LocationScaleEMGradWrapper, LocationEMGradWrapper]:
        loc = self.first_order_moment_num_accum / self.first_order_moment_denom_accum
        if self.scale_trainable:
            scale = tf.sqrt(
                self.second_order_moment_num_accum
                / self.second_order_moment_denom_accum
                - tf.square(loc)
            )
            if self.scale_constraint is not None:
                scale = self.scale_constraint(scale)
            dist = self._build_distribution_from_loc_and_scale(loc=loc, scale=scale)
            return LocationScaleEMGradWrapper(
                dist,
                self.first_order_moment_denom_accum,
                self.first_order_moment_num_accum,
                self.second_order_moment_denom_accum,
                self.second_order_moment_num_accum,
            )
        else:
            scale = self.scale
            if self.scale_constraint is not None:
                scale = self.scale_constraint(scale)
            dist = self._build_distribution_from_loc_and_scale(loc=loc, scale=scale)
            return LocationEMGradWrapper(
                dist,
                self.first_order_moment_denom_accum,
                self.first_order_moment_num_accum,
            )

    def _create_loc_scale_accumulators(self, shape: Tuple[Optional[int], ...]) -> None:
        self.first_order_moment_denom_accum = self.add_weight(
            name="first_order_moment_denom_accum",
            shape=shape,
            initializer=self.accumulator_initializer,
            trainable=self.location_trainable,
        )
        self.first_order_moment_num_accum = self.add_weight(
            name="first_order_moment_num_accum",
            shape=shape,
            initializer=self.location_initializer,
            trainable=self.location_trainable,
        )
        self.first_order_moment_num_accum.assign(
            self.first_order_moment_num_accum * self.first_order_moment_denom_accum
        )

        if self.scale_trainable:
            self.second_order_moment_denom_accum = self.add_weight(
                name="second_order_moment_denom_accum",
                shape=shape,
                initializer=self.accumulator_initializer,
                trainable=True,
            )
            self.second_order_moment_num_accum = self.add_weight(
                name="second_order_moment_num_accum",
                shape=shape,
                initializer=self.scale_initializer,
                trainable=True,
            )
            loc = (
                self.first_order_moment_num_accum / self.first_order_moment_denom_accum
            )
            self.second_order_moment_num_accum.assign(
                (tf.square(self.second_order_moment_num_accum) + tf.square(loc))
                * self.second_order_moment_denom_accum
            )
        else:
            self.scale = self.add_weight(
                name="scale",
                shape=shape,
                initializer=self.scale_initializer,
                trainable=False,
            )

    @abc.abstractmethod
    def _build_distribution_from_loc_and_scale(
        self, loc: tf.Tensor, scale: tf.Tensor
    ) -> tfp.distributions.Distribution:
        pass

    def _create_loc_scale_vars(self, shape: Tuple[Optional[int], ...]) -> None:
        self.loc = self.add_weight(
            name="location",
            shape=shape,
            initializer=self.location_initializer,
            trainable=self.location_trainable,
        )

        if self.scale_trainable:
            self.scale = self.add_weight(
                name="scale",
                shape=shape,
                initializer=self.scale_initializer,
                trainable=self.scale_trainable,
                constraint=GreaterEqualEpsilon(-2.0),
            )
            self.scale.assign(softplus_inverse(self.scale))
        else:
            self.scale = self.add_weight(
                name="scale",
                shape=shape,
                initializer=self.scale_initializer,
                trainable=self.scale_trainable,
            )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(
            scale_initializer=initializers.serialize(self.scale_initializer),
            location_initializer=initializers.serialize(self.location_initializer),
            accumulator_initializer=initializers.serialize(
                self.accumulator_initializer
            ),
            scale_trainable=self.scale_trainable,
            use_accumulators=self.use_accumulators,
            location_trainable=self.location_trainable,
        )
        base_config = super(LocationScaleLeafBase, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def get_modes(self) -> tf.Tensor:
        """
        Obtain the distribution modes.

        This can be used for e.g. MPE estimates of inputs.

        Returns:
            A dict holding the configuration of the layer.
        """
        return self._get_distribution().mode()


class NormalLeaf(LocationScaleLeafBase):
    """
    Computes the log probability of multiple components per variable along the final axis.

    Each component is modelled as a normal distribution with a diagonal covariance matrix.

    Args:
        num_components: Number of components per variable
        location_initializer: Initializer for location variable
        location_trainable: Boolean that indicates whether location is trainable
        scale_initializer: Initializer for scale variable
        scale_trainable: Boolean that indicates whether scale is trainable
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def _build_distribution_from_loc_and_scale(
        self, loc: tf.Tensor, scale: tf.Tensor
    ) -> tfp.distributions.Distribution:
        return tfp.distributions.Normal(loc=loc, scale=scale)


class CauchyLeaf(LocationScaleLeafBase):
    """
    Computes the log probability of multiple components per variable along the final axis.

    Each component is modelled as a Cauchy distribution with a diagonal location matrix.

    Args:
        num_components: Number of components per variable
        location_initializer: Initializer for location variable
        location_trainable: Boolean that indicates whether location is trainable
        scale_initializer: Initializer for scale variable
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def __init__(
        self,
        num_components: int,
        location_initializer: Optional[tf.keras.initializers.Initializer] = None,
        location_trainable: bool = True,
        scale_initializer: Optional[tf.keras.initializers.Initializer] = None,
        accumulator_initializer: Optional[tf.keras.initializers.Initializer] = None,
        use_accumulators: bool = False,
        scale_trainable: bool = False,
        **kwargs
    ):
        super().__init__(
            num_components=num_components,
            location_initializer=location_initializer,
            location_trainable=location_trainable,
            scale_initializer=scale_initializer,
            scale_trainable=scale_trainable,
            accumulator_initializer=accumulator_initializer,
            use_accumulators=use_accumulators,
            **kwargs
        )

    def _build_distribution_from_loc_and_scale(
        self, loc: tf.Tensor, scale: tf.Tensor
    ) -> tfp.distributions.Distribution:
        return tfp.distributions.Cauchy(loc=loc, scale=scale)


class LaplaceLeaf(LocationScaleLeafBase):
    """
    Computes the log probability of multiple components per variable along the final axis.

    Each  component is modelled as a Laplace distribution with a diagonal location matrix.

    Args:
        num_components: Number of components per variable
        location_initializer: Initializer for location variable
        location_trainable: Boolean that indicates whether location is trainable
        scale_initializer: Initializer for scale variable
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def __init__(
        self,
        num_components: int,
        location_initializer: Optional[tf.keras.initializers.Initializer] = None,
        location_trainable: bool = True,
        scale_initializer: Optional[tf.keras.initializers.Initializer] = None,
        accumulator_initializer: Optional[tf.keras.initializers.Initializer] = None,
        use_accumulators: bool = False,
        **kwargs
    ):
        super().__init__(
            num_components=num_components,
            location_initializer=location_initializer,
            location_trainable=location_trainable,
            scale_initializer=scale_initializer,
            scale_trainable=False,
            accumulator_initializer=accumulator_initializer,
            use_accumulators=use_accumulators,
            **kwargs
        )

    def _build_distribution_from_loc_and_scale(
        self, loc: tf.Tensor, scale: tf.Tensor
    ) -> tfp.distributions.Distribution:
        return tfp.distributions.Laplace(loc=loc, scale=scale)



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras


class FlatToRegions(keras.layers.Layer):
    """
    Flat representation to a dense representation.

    Reshapes a flat input of shape ``[batch, num_vars[, var_dimensionality]]``
    to ``[batch, num_vars == scopes, decomp, var_dimensionality]``

    If ``var_dimensionality`` is 1, the shape can also be ``[batch, num_vars]``.

    Args:
        **kwargs: Keyword arguments to pass on the keras.Layer super class
    """

    def __init__(self, num_decomps: int, **kwargs):
        self.num_decomps = num_decomps
        super(FlatToRegions, self).__init__(**kwargs)

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Reshape to a region representation.

        Args:
            inputs: Flat representation of raw input to the SPN.
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with axes ``[batch, scopes, decomps, nodes]``.
        """
        if len(inputs.shape) == 2:
            inputs = tf.expand_dims(inputs, axis=-1)
        with_decomps = tf.tile(
            tf.expand_dims(inputs, axis=2), multiples=(1, 1, self.num_decomps, 1)
        )
        return with_decomps

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.
        """
        num_batch, num_vars, var_dimensionality = input_shape
        return num_batch, num_vars, self.num_decomps, var_dimensionality

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(num_decomps=self.num_decomps)
        base_config = super(FlatToRegions, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Callable, Optional, Tuple

import tensorflow as tf
from tensorflow import initializers


def logspace_wrapper_initializer(
    initializer: initializers.Initializer,
) -> Callable[[Tuple[Optional[int], ...], tf.dtypes.DType], tf.Tensor]:
    """
    Wrap an initializer so that its values are projected in log-space.

    Args:
        initializer: The initializer to convert to logspace

    Returns:
        A initialization callable that produces the log-space representation of `initializer`
    """

    def _wrap_fn(
        shape: Tuple[Optional[int], ...], dtype: tf.dtypes.DType = None
    ) -> tf.Tensor:
        return tf.math.log(initializer(shape=shape, dtype=dtype))

    return _wrap_fn



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.utils import tf_utils


class LogDropout(keras.layers.Layer):
    """
    Log dropout layer.

    Applies dropout in log-space. Should not precede product layers in an SPN, since their scope probability
    then potentially becomes -inf, resulting in NaN-values during training.

    Args:
        rate: Rate at which to randomly dropout inputs.
        noise_shape: Shape of dropout noise tensor
        seed: Random seed
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def __init__(
        self,
        rate: float,
        noise_shape: Optional[Tuple[int, ...]] = None,
        seed: int = None,
        axis_at_least_one: Optional[int] = None,
        **kwargs
    ):

        super(LogDropout, self).__init__(**kwargs)
        self.rate = min(1.0, max(0.0, rate))
        self.noise_shape = noise_shape
        self.seed = seed
        self.supports_masking = True
        self.axis_at_least_one = axis_at_least_one

    def _get_noise_shape(self, inputs: tf.Tensor) -> tf.Tensor:
        if self.noise_shape is None:
            return tf.shape(inputs)

        return tf.concat([[tf.shape(inputs)[0]], self.noise_shape], axis=0)

    def call(self, inputs: tf.Tensor, training: Optional[bool] = None) -> tf.Tensor:
        """
        Randomly drop out parts of incoming tensor by setting the log probability to -Inf.

        Args:
            inputs: Input log-probability Tensor.
            training: Whether or not the network is training.

        Returns:
            Tensor with some nodes dropped out.
        """
        if training is None:
            training = keras.backend.learning_phase()

        def dropped_inputs() -> tf.Tensor:
            noise_tensor = tf.random.uniform(
                shape=self._get_noise_shape(inputs), seed=self.seed
            )
            keep_tensor = tf.greater(noise_tensor, self.rate)
            if self.axis_at_least_one is not None:
                keep_tensor = tf.logical_or(
                    keep_tensor,
                    tf.equal(
                        noise_tensor,
                        tf.reduce_max(
                            noise_tensor, axis=self.axis_at_least_one, keepdims=True
                        ),
                    ),
                )
            return tf.reshape(
                tf.where(keep_tensor, inputs, float("-inf")), shape=tf.shape(inputs)
            )

        if self.rate == 0.0:
            return tf.identity(inputs)

        output = tf_utils.smart_cond(
            training, dropped_inputs, lambda: tf.identity(inputs)
        )
        return output

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(rate=self.rate, noise_shape=self.noise_shape, seed=self.seed)
        base_config = super(LogDropout, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.
        """
        return input_shape



from libspn_keras.layers.base_leaf import BaseLeaf
from libspn_keras.layers.conv2d_product import Conv2DProduct
from libspn_keras.layers.conv2d_sum import Conv2DSum
from libspn_keras.layers.dense_product import DenseProduct
from libspn_keras.layers.dense_sum import DenseSum
from libspn_keras.layers.flat_to_regions import FlatToRegions
from libspn_keras.layers.indicator_leaf import IndicatorLeaf
from libspn_keras.layers.local2d_sum import Local2DSum
from libspn_keras.layers.location_scale_leaf import (
    CauchyLeaf,
    LaplaceLeaf,
    LocationScaleLeafBase,
    NormalLeaf,
)
from libspn_keras.layers.log_dropout import LogDropout
from libspn_keras.layers.normalize_standard_score import NormalizeStandardScore
from libspn_keras.layers.permute_and_pad_scopes import PermuteAndPadScopes
from libspn_keras.layers.permute_and_pad_scopes_random import PermuteAndPadScopesRandom
from libspn_keras.layers.reduce_product import ReduceProduct
from libspn_keras.layers.root_sum import RootSum
from libspn_keras.layers.spatial_to_regions import SpatialToRegions
from libspn_keras.layers.temporal_dense_product import TemporalDenseProduct
from libspn_keras.layers.undecompose import Undecompose

__all__ = [
    "Conv2DProduct",
    "DenseProduct",
    "DenseSum",
    "IndicatorLeaf",
    "NormalLeaf",
    "LaplaceLeaf",
    "CauchyLeaf",
    "LocationScaleLeafBase",
    "LogDropout",
    "SpatialToRegions",
    "RootSum",
    "Local2DSum",
    "Undecompose",
    "NormalizeStandardScore",
    "BaseLeaf",
    "PermuteAndPadScopes",
    "FlatToRegions",
    "ReduceProduct",
    "PermuteAndPadScopesRandom",
    "TemporalDenseProduct",
    "Conv2DSum",
]



import logging
from typing import List, Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras import initializers

from libspn_keras.layers.permute_and_pad_scopes import PermuteAndPadScopes


logger = logging.getLogger("libspn-keras")


class PermuteAndPadScopesRandom(PermuteAndPadScopes):
    """
    Permutes scopes, usually applied after a ``FlatToRegions`` and a ``BaseLeaf`` layer.

    Args:
        factors: Number of factors in preceding product layers. Needed to compute
            the effective number of scopes, including padded nodes. Can be applied at later stage
            through ``generate_factors``.
        **kwargs: kwargs to pass on to the ``keras.Layer`` superclass.
    """

    def __init__(self, factors: Optional[List[int]] = None, **kwargs):
        super(PermuteAndPadScopesRandom, self).__init__(None, **kwargs)
        self.factors = factors

    def set_factors(self, factors: List[int]) -> None:
        """
        Set the factors.

        These factors determine the amount of padding which is required to for uniform dimension sizes
        across scope, decomp and node axes.

        Args:
            factors: The factors of products in the SPN graph per layer (DenseProduct or ReduceProduct).
        """
        self.factors = factors

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:  # noqa: C901
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: If shape cannot be determined
        """
        _, num_scopes, num_decomps, num_nodes_in = input_shape
        if num_decomps is None:
            raise ValueError(
                "Cannot build PermuteAndPadScopesRandom with an unknown decomposition dimension"
            )
        if num_scopes is None:
            raise ValueError(
                "Cannot build PermuteAndPadScopesRandom with an unknown scopes dimension"
            )
        if num_nodes_in is None:
            raise ValueError(
                "Cannot build PermuteAndPadScopesRandom with an unknown number of nodes per "
                "scope/decomposition"
            )

        if self.factors is None or self.factors == []:
            raise ValueError("Factors needs to be a non-empty sequence.")
        factor_cumprod = np.cumprod(self.factors)
        factor_prod = factor_cumprod[-1]
        if factor_prod < num_scopes:
            raise ValueError(
                "{}: not enough factors to cover all variables ({} vs. {}).".format(
                    self, factor_prod, num_scopes
                )
            )
        for i, fc in enumerate(factor_cumprod[:-1]):
            if fc >= num_scopes:
                raise ValueError(
                    "{}: too many factors, taking out the bottom {} products still "
                    "results in {} factors while {} are needed.".format(
                        self, len(self.factors) - i - 1, fc, num_scopes
                    )
                )

        # Now we generate the random index permutations
        perms: List[List[int]] = [
            list(np.random.permutation(num_scopes).astype(int))
            for _ in range(num_decomps)
        ]

        num_m1 = factor_prod - num_scopes
        if num_m1 > 0:
            # e.g. num_m1 == 2 and factor_prod = 32. Then rate_m1 is 16, so once every 16 values
            # we should leave a variable slot empty
            rate_m1 = int(np.floor(factor_prod / num_m1))

            for p in perms:
                for i in range(num_m1):
                    p.insert(i * rate_m1, -1)
        self.permutations = self.add_weight(
            name="permutations",
            initializer=initializers.Constant(perms),
            trainable=False,
            shape=[num_decomps, factor_prod],
            dtype=tf.int32,
        )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(num_vars_input=self.num_vars_spn_input, factors=self.factors)
        base_config = super(PermuteAndPadScopesRandom, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from .clip import Clip
from .greater_equal_epsilon import GreaterEqualEpsilon
from .greater_equal_epsilon_normalized import GreaterEqualEpsilonNormalized
from .log_normalized import LogNormalized

__all__ = [
    "Clip",
    "GreaterEqualEpsilon",
    "GreaterEqualEpsilonNormalized",
    "LogNormalized",
]



from typing import Optional, Tuple

from libspn_keras.layers.dense_sum import DenseSum
from libspn_keras.logspace import logspace_wrapper_initializer


class Local2DSum(DenseSum):
    """
    Computes a spatial local sum, i.e. all cells will have unique weights.

    In other words, there is no weight sharing across the spatial axes.

    Args:
        num_sums: Number of sums per spatial cell. Corresponds to the number of channels in
            the output
        logspace_accumulators: If ``True``, accumulators will be represented in log-space which
            is typically used with ``BackpropMode.GRADIENT``. If ``False``, accumulators will be
            represented in linear space. Weights are computed by normalizing the accumulators
            per sum, so that we always end up with a normalized SPN. If ``None`` (default) it
            will be set to ``True`` for ``BackpropMode.GRADIENT`` and ``False`` otherwise.
        accumulator_initializer: Initializer for accumulator
        accumulator_regularizer: Regularizer for accumulators
        linear_accumulator_constraint: Constraint for accumulators (only applied if
            log_space_accumulators==False)
        sum_op (SumOpBase): SumOpBase instance which determines how to compute the forward and
            backward pass of the weighted sums
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.
        """
        # Create a trainable weight variable for this layer.
        _, num_scopes_vertical, num_scopes_horizontal, num_channels_in = input_shape

        weights_shape = (
            num_scopes_vertical,
            num_scopes_horizontal,
            num_channels_in,
            self.num_sums,
        )

        initializer = self.accumulator_initializer
        accumulator_contraint = self.linear_accumulator_constraint
        if self.logspace_accumulators:
            initializer = logspace_wrapper_initializer(initializer)
            accumulator_contraint = self.logspace_accumulator_constraint

        self._accumulators = self.add_weight(
            name="sum_weights",
            shape=weights_shape,
            initializer=initializer,
            regularizer=self.accumulator_regularizer,
            constraint=accumulator_contraint,
        )
        super(Local2DSum, self).build(input_shape)



import functools
import operator
from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras


class DenseProduct(keras.layers.Layer):
    """
    Computes products per decomposition and scope by an 'n-order' outer product.

    Assumes the incoming tensor is of shape ``[num_scopes, num_decomps, num_batch, num_nodes]`` and
    produces an output of ``[num_scopes // num_factors, num_decomps, num_batch, num_nodes ** num_factors]``.
    It can be considered a *dense* product as it computes all possible products given the scopes it has
    to merge.

    Args:
        num_factors (int): Number of factors per product
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def __init__(self, num_factors: int, **kwargs):
        super(DenseProduct, self).__init__(**kwargs)
        self.num_factors = num_factors

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: When shape could not be determined.
        """
        _, self._num_scopes_in, self._num_decomps, self._num_nodes_in = input_shape
        if self._num_scopes_in is None:
            raise ValueError("Cannot build with unknown number of input scopes")
        if self._num_nodes_in is None:
            raise ValueError("Cannot build with unknown number of input nodes")
        if self._num_scopes_in % self.num_factors != 0:
            raise ValueError("Number of input scopes is not divisible by factor")
        self._num_scopes_out = self._num_scopes_in // self.num_factors
        self._num_products = self._num_nodes_in ** self.num_factors
        super(DenseProduct, self).build(input_shape)

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute a dense product layer by using all possible combinations of children.

        Args:
            x: Region Tensor.
            kwargs: Remaining keyword arguments.

        Returns:
            A Tensor with dense products.
        """
        # Split in list of tensors which will be added up using outer products
        shape = [
            -1,
            self._num_scopes_out,
            self.num_factors,
            self._num_decomps,
            self._num_nodes_in,
        ]
        with tf.name_scope("LogProbPerFactor"):
            log_prob_per_factor = tf.split(
                tf.reshape(x, shape=shape), axis=2, num_or_size_splits=self.num_factors
            )

        # Reshape to [scopes, decomps, batch, 1, ..., child.dim_nodes, ..., 1] where
        # child.dim_nodes is inserted at the i-th index within the trailing 1s, where i corresponds
        # to the index of the log prob
        with tf.name_scope("ToBroadcastable"):
            log_prob_per_factor_broadcastable = [
                tf.reshape(
                    log_prob,
                    [-1, self._num_scopes_out, self._num_decomps]
                    + [
                        1 if j != i else self._num_nodes_in
                        for j in range(self.num_factors)
                    ],
                )
                for i, log_prob in enumerate(log_prob_per_factor)
            ]
        # Add up everything (effectively computing an outer product) and flatten the last
        # num_factors dimensions.
        with tf.name_scope("NOrderOuterProduct"):
            outer_product = functools.reduce(
                operator.add, log_prob_per_factor_broadcastable
            )
            return tf.reshape(
                outer_product,
                [-1, self._num_scopes_out, self._num_decomps, self._num_products],
            )

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.

        Raises:
            ValueError: When shape cannot be determined.
        """
        num_batch, num_scopes_in, num_decomps, num_nodes_in = input_shape
        if num_scopes_in is None:
            raise ValueError(
                "Cannot compute output shape with unknown number of input scopes"
            )
        if num_nodes_in is None:
            raise ValueError(
                "Cannot compute output shape with unknown number of input nodes"
            )
        return (
            num_batch,
            num_scopes_in // self.num_factors,
            self._num_decomps,
            int(num_nodes_in ** self.num_factors),
        )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(num_factors=self.num_factors,)
        base_config = super(DenseProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

from libspn_keras import config
from libspn_keras import constraints
from libspn_keras import initializers
from libspn_keras import layers
from libspn_keras import losses
from libspn_keras import metrics
from libspn_keras import models
from libspn_keras import optimizers
from libspn_keras.config.accumulator_initializer import (
    get_default_accumulator_initializer,
    set_default_accumulator_initializer,
)
from libspn_keras.config.linear_accumulator_constraint import (
    get_default_linear_accumulators_constraint,
    set_default_linear_accumulators_constraint,
)
from libspn_keras.config.logspace_accumulator_constraint import (
    get_default_logspace_accumulators_constraint,
    set_default_logspace_accumulators_constraint,
)
from libspn_keras.config.sum_op import get_default_sum_op, set_default_sum_op
from libspn_keras.logspace import logspace_wrapper_initializer
from libspn_keras.region import region_graph_to_dense_spn
from libspn_keras.region import RegionNode
from libspn_keras.region import RegionVariable
from libspn_keras.sum_ops import (
    SumOpEMBackprop,
    SumOpGradBackprop,
    SumOpHardEMBackprop,
    SumOpUnweightedHardEMBackprop,
)
from libspn_keras.visualize import visualize_dense_spn


__all__ = [
    "config",
    "get_default_accumulator_initializer",
    "set_default_accumulator_initializer",
    "set_default_logspace_accumulators_constraint",
    "set_default_linear_accumulators_constraint",
    "get_default_linear_accumulators_constraint",
    "get_default_logspace_accumulators_constraint",
    "get_default_sum_op",
    "set_default_sum_op",
    "logspace_wrapper_initializer",
    "optimizers",
    "metrics",
    "losses",
    "layers",
    "constraints",
    "RegionNode",
    "RegionVariable",
    "region_graph_to_dense_spn",
    "visualize_dense_spn",
    "initializers",
    "models",
    "SumOpUnweightedHardEMBackprop",
    "SumOpEMBackprop",
    "SumOpGradBackprop",
    "SumOpHardEMBackprop",
]



import tensorflow as tf
from tensorflow.keras.constraints import Constraint


class GreaterEqualEpsilonNormalized(Constraint):
    """
    Constraints the weight to be greater than or equal to epsilon and then normalizes.

    Args:
        epsilon: Constant, usually small non-zero
    """

    def __init__(self, epsilon: float = 1e-10, axis: int = -2):
        self.epsilon = epsilon
        self.axis = axis

    def __call__(self, w: tf.Tensor) -> tf.Tensor:
        """
        Lower-clip input tensor to ``epsilon`` and then normalize.

        Args:
            w: Weight Tensor.

        Returns:
            Clipped weight Tensor.
        """
        clipped = tf.maximum(
            w, self.epsilon / tf.cast(tf.shape(w)[self.axis], tf.float32)
        )
        return clipped / tf.reduce_sum(clipped, axis=self.axis, keepdims=True)

    def get_config(self) -> dict:
        """
        Obtain config.

        Returns:
            Key-value mapping with configuration of this constraint.
        """
        return dict(epsilon=self.epsilon, axis=self.axis)



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow.keras import initializers


class EpsilonInverseFanIn(initializers.Initializer):
    r"""
    Initializes all values in a tensor with :math:`\epsilon K^{-1}`.

    Where :math:`K` is the dimension at ``axis``.

    This is particularly useful for (unweighted) hard EM learning and should generally be avoided
    otherwise.

    Args:
        axis: The axis for input nodes so that :math:`K^{-1}` is the inverse fan in. Usually,
            this is ``-2``.
        epsilon: A small non-zero constant
    """

    def __init__(
        self, axis: int = -2, epsilon: float = 1e-4,
    ):
        self.axis = axis
        self.epsilon = epsilon

    def __call__(
        self, shape: Tuple[int, ...], dtype: Optional[tf.DType] = None
    ) -> tf.Tensor:
        """
        Compute initializations along the given axis.

        Args:
            shape: Shape of Tensor to initialize.
            dtype: DType of Tensor to initialize.

        Returns:
            Initial value.
        """
        n = shape[self.axis]
        return tf.cast(tf.fill(shape, self.epsilon / n), dtype)

    def get_config(self) -> dict:
        """
        Obtain the config.

        Returns:
            Key-value mapping with the configuration of this initializer.
        """
        return {"epsilon": self.epsilon, "axis": self.axis}
